#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for making plots for background classifiers (muon/random_noise/neutrinos).
"""

from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
from orcanet_contrib.plotting.utils import get_mc_info_and_other_datasets


def make_prob_hists_bg_classifier(pred_file, savefolder, bg_classes=('not_neutrino', 'neutrino'),
                                  cuts=None, savename_prefix=None, x_ranges=((0, 1),), xlabel_prefix='OrcaNet'):
    """
    Function that makes plots for the reco probability distributions of the background classifier.

    Makes one plot for each probability output (prob_muon, prob_random_noise, prob_neutrino)

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the background classification predictions of a nn model.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    bg_classes : tpl(str)
        Tuple that specifies the single classes of a background classifier.
        This is needed to find the arrays for the class probabilities in the pred dataset.
        I.e. if you supply bg_classes=('not_neutrino', 'neutrino'), the pred dataset should
        contain the rows "pred_not_neutrino" and "pred_neutrino".
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.
    savename_prefix : None/str
        Optional string prefix for the savename of the plot.
    x_ranges : list(tpl)
        List of tuples with various x_ranges that should be made of the same plot.
    xlabel_prefix : str
        Prefix that is in front of the x_axis label.

    """
    def configure_and_save_plot():
        """
        Configure and save a mpl plot with GridLines, Logscale etc. for the background classifier probability plots.
        """
        axes.legend(loc='upper center', ncol=1)
        plt.grid(True, zorder=0, linestyle='dotted')
        plt.yscale('log')

        plt.minorticks_on()

        plt.xlabel(xlabel_prefix + ' ' + bg_class + ' probability')
        plt.ylabel('Normed Quantity')
        title = plt.title('Probability to be classified as ' + bg_class)
        title.set_position([.5, 1.04])

        pdf_plots.savefig(fig)

    dsets = get_mc_info_and_other_datasets(pred_file, 'mc_info', 'pred', cuts=cuts)
    mc_info, pred = dsets[0], dsets[1]
    ptype, is_cc = mc_info['particle_type'], mc_info['is_cc']

    for bg_class in bg_classes:
        prob_class = pred['prob_' + bg_class]

        savename = 'prob_' + bg_class if savename_prefix is None else savename_prefix + '_prob_' + bg_class
        pdf_plots = PdfPages(savefolder + '/' + savename + '.pdf')

        for x_range in x_ranges:
            fig, axes = plt.subplots()
            make_prob_hists_for_class(prob_class, ptype, is_cc, axes, x_range=x_range)
            configure_and_save_plot()
            plt.close()

        pdf_plots.close()


def select_class(class_name, ptype, is_cc):
    """
    Returns a boolean array which specifies, which rows in the ptype & is_cc 1d arrays belong to the class.

    Parameters
    ----------
    class_name : str
        String that specifies the class that should be selected for the boolean flag output.
    ptype : ndarray(ndim=1)
        Array with particle_types of some events.
    is_cc : ndarray(ndim=1)
        Array with is_cc of some events.

    Returns
    -------
    is_class : ndarray(ndim=1)
        Boolean flag, which specifies if each row belongs to the class specified by the class_name.

    """
    if class_name == 'mupage':
        is_class = np.abs(ptype) == 13
    elif class_name == 'random_noise':
        is_class = np.abs(ptype) == 0
    elif class_name == 'neutrino':
        is_class = np.logical_or.reduce((np.abs(ptype) == 12, np.abs(ptype) == 14, np.abs(ptype) == 16))
    elif class_name == 'neutrino_track':
        is_class = np.logical_and(np.abs(ptype) == 14, is_cc == 1)
    elif class_name == 'neutrino_shower':
        is_muon_nc = np.logical_and(np.abs(ptype) == 14, is_cc == 0)
        is_class = np.logical_or.reduce((np.abs(ptype) == 12, np.abs(ptype) == 16, is_muon_nc))
    else:
        raise ValueError('The class ' + str(class_name) + ' is not available.')

    return is_class


def make_prob_hists_for_class(prob_class, ptype, is_cc, axes, x_range=(0, 1)):
    """
    Makes a mpl hist plot of the probabilities for a certain probability class (either muon, random_noise or neutrino).

    Plots the probabilities for this prob_class for all events (muon, random_noise, neutrinos)

    Parameters
    ----------
    prob_class : ndarray(ndim=1)
        Reco probabilities for events for a certain probability class (muon, random_noise, neutrinos).
    ptype : ndarray(ndim=1)
        Array with particle_types of the same events as in the prob_class.
    is_cc : ndarray(ndim=1)
        Array with is_cc of the same events as in the prob_class.
    axes : mpl.axes
        Axes object that refers to ax of an existing plt.sublots object.
    x_range : tpl
        Sets the x_range of the prob hist plot.

    """
    is_mupage = select_class('mupage', ptype, is_cc)
    is_random_noise = select_class('random_noise', ptype, is_cc)
    is_neutrino = select_class('neutrino', ptype, is_cc)

    if x_range != (0, 1):
        x_range_diff = x_range[1] - x_range[0]
        bins = 100 * int(1 / x_range_diff)
    else:
        bins = 100

    axes.hist(prob_class[is_mupage], density=True, bins=bins, range=(0, 1), color='b', label='mupage', histtype='step', linestyle='-', zorder=3)
    axes.hist(prob_class[is_random_noise], density=True, bins=bins, range=(0, 1), color='r', label='random_noise', histtype='step', linestyle='-', zorder=3)
    axes.hist(prob_class[is_neutrino], density=True, bins=bins, range=(0, 1), color='saddlebrown', label='neutrino', histtype='step', linestyle='-', zorder=3)

    if x_range != (0, 1):
        axes.set_xlim(x_range)


def make_contamination_to_neutrino_efficiency_plot(pred_file_1, pred_file_2, dataset_modifier, savefolder, cuts=None):
    """
    Makes bg classifier plots like muon contamination vs neutrino efficiency, both weighted 1 year and non weighted.

    Parameters
    ----------
    pred_file_1 : h5py.File
        H5py file instance, which stores the background classification predictions of a nn model.
    pred_file_2 : h5py.File
        H5py file instance of a second reco result, typically the results of the standard reconstruction.
    dataset_modifier : str
        Specifies which dataset_modifier has been used in creating both pred files.
    savefolder : str
        Full dirpath of the folder where all the plots of this function should be saved.

    """
    mc_info_1, mc_info_2 = pred_file_1['mc_info'], pred_file_2['mc_info']
    pred_1, pred_2 = pred_file_1['pred'], pred_file_2['pred']

    if cuts is not None:
        assert type(cuts) is tuple
        mc_info_1, mc_info_2 = mc_info_1[cuts[0]], mc_info_2[cuts[1]]
        pred_1, pred_2 = pred_1[cuts[0]], pred_2[cuts[1]]

    if dataset_modifier == 'bg_classifier_2_class':
        pdf_plots = PdfPages(savefolder + '/muon_contamination_to_neutr_efficiency.pdf')
        plot_contamination_to_neutr_eff_multi_e_cut(mc_info_1, mc_info_2, pred_1, pred_2, pdf_plots)

        pdf_plots.close()


def plot_contamination_to_neutr_eff_multi_e_cut(mc_info_dl, mc_info_std, pred_dl, pred_std, pdf_plots,
                                                overlay=('KM3NeT Preliminary', (0.2, 0.7))):
    """
    Make muon/random-noise contamination plots vs neutrino efficiency.

    TODO improve readability of function.

    Parameters
    ----------
    mc_info_dl : ndarray(ndim=2)
        Structured array containing the MC info of reco 1.
    mc_info_std : ndarray(ndim=2)
        Structured array containing the MC info of reco 2 (typically standard reco).
    pred_dl : ndarray(ndim=2)
        Structured array containing the pred_dl info of reco 1.
    pred_std : ndarray(ndim=2)
        Structured array containing the pred_dl info of reco 2 (typically standard reco).
    pdf_plots : mpl.backends.backend_pdf.PdfPages
        Matplotlib pdfpages instance, to which the plots should be saved.

    """
    def make_plots():
        # -- Make weighted plots -- #

        # - Make continous plots
        # Muon contamination

        fig, ax = plt.subplots()

        ax.plot(np.array(muon_cont_weighted_dl), np.array(neutrino_eff_weighted_dl), label='CNN')
        ax.plot(np.array(muon_cont_weighted_std), np.array(neutrino_eff_weighted_std), label='RF')

        ax.set_xlim(left=0, right=20)
        ax.set_xlabel('Muon contamination [%]'), ax.set_ylabel('Neutrino Efficiency [%]')
        ax.grid(True)
        ax.legend(loc='center right')
        plt.title('E-range:' + str(tpl) + ', weighted for an atmospheric flux')

        pdf_plots.savefig(fig)
        plt.close()

        # Random noise contamination
        fig, ax = plt.subplots()

        ax.plot(np.array(rn_cont_weighted_dl), np.array(neutrino_eff_weighted_dl), label='CNN')
        ax.plot(np.array(rn_cont_weighted_std), np.array(neutrino_eff_weighted_std), label='RF')

        ax.set_xlim(left=0, right=5)
        ax.set_xlabel('Random noise contamination [%]'), ax.set_ylabel('Neutrino Efficiency [%]')
        ax.grid(True)
        ax.legend(loc='center right')
        plt.title('E-range:' + str(tpl) + ', weighted for 10kHz noise')

        pdf_plots.savefig(fig)
        plt.close()

        # - Make errorbar plots
        # Muon contamination

        fig, ax = plt.subplots()

        data_points_range_mupage = np.linspace(0, 20, 21)
        idx_mupage_closest_dl, idx_mupage_closest_std = [], []
        for j in range(len(data_points_range_mupage)):
            dpoint = data_points_range_mupage[j]

            idx_mupage_closest_dl.append((np.abs(np.array(muon_cont_weighted_dl) - dpoint)).argmin())
            idx_mupage_closest_std.append((np.abs(np.array(muon_cont_weighted_std) - dpoint)).argmin())

        dpoints_neutr_eff_dl = np.array(neutrino_eff_weighted_dl)[idx_mupage_closest_dl]
        dpoints_muon_cont_dl = np.array(muon_cont_weighted_dl)[idx_mupage_closest_dl]
        dpoints_muon_cont_err_dl = np.array(muon_cont_weighted_dl_err)[idx_mupage_closest_dl]

        dpoints_neutr_eff_std = np.array(neutrino_eff_weighted_std)[idx_mupage_closest_std]
        dpoints_muon_cont_std = np.array(muon_cont_weighted_std)[idx_mupage_closest_std]
        dpoints_muon_cont_err_std = np.array(muon_cont_weighted_std_err)[idx_mupage_closest_std]

        plt.errorbar(dpoints_muon_cont_dl, dpoints_neutr_eff_dl, xerr=dpoints_muon_cont_err_dl, fmt='.', markersize=4, label='CNN', linewidth=0.7)
        plt.errorbar(dpoints_muon_cont_std, dpoints_neutr_eff_std, xerr=dpoints_muon_cont_err_std, fmt='.', markersize=4, label='RF', linewidth=0.7)

        ax.set_xlim(left=0, right=20)
        ax.set_xlabel('Muon contamination [%]'), ax.set_ylabel('Neutrino Efficiency [%]')
        ax.grid(True)
        ax.legend(loc='center right')
        plt.title('E-range:' + str(tpl) + ', weighted for an atmospheric flux')
        plt.text(overlay[1][0], overlay[1][1], overlay[0], transform=ax.transAxes, weight='bold')

        pdf_plots.savefig(fig)

        # Just because we want a plot from 85 to 105
        ax.set_ylim(bottom=85, top=ax.get_ylim()[1])
        pdf_plots.savefig(fig)

        plt.close()

        # Random noise contamination

        fig, ax = plt.subplots()

        data_points_range_rn = np.linspace(0, 10, 21)
        idx_rn_closest_dl, idx_rn_closest_std = [], []
        for j in range(len(data_points_range_rn)):
            dpoint = data_points_range_rn[j]

            idx_rn_closest_dl.append((np.abs(np.array(rn_cont_weighted_dl) - dpoint)).argmin())
            idx_rn_closest_std.append((np.abs(np.array(rn_cont_weighted_std) - dpoint)).argmin())

        dpoints_neutr_eff_dl = np.array(neutrino_eff_weighted_dl)[idx_rn_closest_dl]
        dpoints_rn_cont_dl = np.array(rn_cont_weighted_dl)[idx_rn_closest_dl]
        dpoints_rn_cont_err_dl = np.array(rn_cont_weighted_dl_err)[idx_rn_closest_dl]

        dpoints_neutr_eff_std = np.array(neutrino_eff_weighted_std)[idx_rn_closest_std]
        dpoints_rn_cont_std = np.array(rn_cont_weighted_std)[idx_rn_closest_std]
        dpoints_rn_cont_err_std = np.array(rn_cont_weighted_std_err)[idx_rn_closest_std]

        plt.errorbar(dpoints_rn_cont_dl, dpoints_neutr_eff_dl, xerr=dpoints_rn_cont_err_dl, fmt='.', markersize=4, label='CNN', linewidth=0.7)
        plt.errorbar(dpoints_rn_cont_std, dpoints_neutr_eff_std, xerr=dpoints_rn_cont_err_std, fmt='.', markersize=4, label='RF', linewidth=0.7)

        ax.set_xlim(left=0, right=5)
        ax.set_xlabel('Random noise contamination [%]'), ax.set_ylabel('Neutrino Efficiency [%]')
        ax.grid(True)
        ax.legend(loc='center right')
        plt.title('E-range:' + str(tpl) + ', weighted for 10kHz noise')
        plt.text(overlay[1][0], overlay[1][1], overlay[0], transform=ax.transAxes, weight='bold')

        pdf_plots.savefig(fig)
        plt.close()

    ptype_dl, is_cc_dl = mc_info_dl['particle_type'], mc_info_dl['is_cc']
    ptype_std, is_cc_std = mc_info_std['particle_type'], mc_info_std['is_cc']
    is_neutrino_dl, is_neutrino_std = select_class('neutrino', ptype_dl, is_cc_dl), select_class('neutrino', ptype_std, is_cc_std)
    is_mupage_dl, is_mupage_std = select_class('mupage', ptype_dl, is_cc_dl), select_class('mupage', ptype_std, is_cc_std)
    is_rn_dl, is_rn_std = select_class('random_noise', ptype_dl, is_cc_dl), select_class('random_noise', ptype_std, is_cc_std)

    n_neutrinos_total = np.count_nonzero(is_neutrino_dl)  # dl or std doesnt matter for n_neutrinos_total
    n_muons_total = np.count_nonzero(is_mupage_dl)
    n_rn_total = np.count_nonzero(is_rn_dl)
    print('Total number of neutrinos: ' + str(n_neutrinos_total) + ', ' + str(np.count_nonzero(is_neutrino_std)))
    print('Total number of mupage muons: ' + str(n_muons_total) + ', ' + str(np.count_nonzero(is_mupage_std)))
    print('Total number of random_noise: ' + str(n_rn_total) + ', ' + str(np.count_nonzero(is_rn_std)))

    # define e_cut and cut values on neutrino prob
    e_cuts = ((1, 100), (1, 5), (5, 10), (10, 20), (20, 100))
    cuts = np.logspace(-5.9, 1, num=10000)  # cuts = np.linspace(0, 1, 10000)
    prob_not_neutrino = pred_dl['prob_not_neutrino']
    muon_score_std = pred_std['prob_muon']
    rn_noise_score_std = pred_std['prob_noise']
    w_1_y = mc_info_dl['oscillated_weight_one_year_bg_sel']
    w_1_y_std = mc_info_std['oscillated_weight_one_year_bg_sel']

    for tpl in e_cuts:
        print('Making contamination plots for e-cut ' + str(tpl))

        e_low, e_high = tpl[0], tpl[1]
        e_cut_mask_dl = np.logical_and(mc_info_dl['energy'] >= e_low, mc_info_dl['energy'] < e_high)
        e_cut_mask_std = np.logical_and(mc_info_std['energy'] >= e_low, mc_info_std['energy'] < e_high)

        # make energy cuts on neutrinos only
        is_neutrino_e_cut_dl = np.logical_and(e_cut_mask_dl, is_neutrino_dl)
        is_neutrino_e_cut_std = np.logical_and(e_cut_mask_std, is_neutrino_std)

        # n_neutrinos_total_e_cut = np.count_nonzero(is_neutrino_e_cut_dl)  # should be same for both dl and std
        n_neutrinos_total_e_cut_weighted = np.sum(w_1_y[is_neutrino_e_cut_dl])

        n_neutrinos_total_e_cut_weighted_std = np.sum(w_1_y_std[is_neutrino_e_cut_std])
        assert np.round(n_neutrinos_total_e_cut_weighted_std, 5) == np.round(n_neutrinos_total_e_cut_weighted, 5)

        # x_contamination_mupage_dl, y_neutrino_efficiency_dl, y_n_muons_dl = [], [], []
        # x_contamination_mupage_std, y_neutrino_efficiency_std, y_n_muons_std = [], [], []

        neutrino_eff_weighted_dl, muon_cont_weighted_dl, rn_cont_weighted_dl = [], [], []
        neutrino_eff_weighted_std, muon_cont_weighted_std, rn_cont_weighted_std = [], [], []

        muon_cont_weighted_dl_err, muon_cont_weighted_std_err = [], []
        rn_cont_weighted_dl_err, rn_cont_weighted_std_err = [], []

        for i in range(len(cuts)):

            # dl

            # if np.count_nonzero(mupage_sel_dl) != 0:
            #     fract_n_mupage_err_to_n_mupage_dl = n_mupage_err_dl / np.count_nonzero(mupage_sel_dl)
            # else:
            #     fract_n_mupage_err_to_n_mupage_dl = 0

            neutr_sel_dl = prob_not_neutrino[is_neutrino_e_cut_dl] < cuts[i]
            n_neutrino_weighted_dl = np.sum(w_1_y[is_neutrino_e_cut_dl][neutr_sel_dl])

            if n_neutrino_weighted_dl != 0:
                mupage_sel_dl = prob_not_neutrino[is_mupage_dl] < cuts[i]
                rn_sel_dl = prob_not_neutrino[is_rn_dl] < cuts[i]

                n_mupage_weighted_dl = np.sum(w_1_y[is_mupage_dl][mupage_sel_dl])
                n_rn_weighted_dl = np.sum(w_1_y[is_rn_dl][rn_sel_dl])

                neutrino_eff_weighted_dl.append((n_neutrino_weighted_dl / n_neutrinos_total_e_cut_weighted) * 100)
                muon_cont_weighted_dl.append((n_mupage_weighted_dl / n_neutrino_weighted_dl) * 100)
                rn_cont_weighted_dl.append((n_rn_weighted_dl / n_neutrino_weighted_dl) * 100)

                # we calculate n_mupage_weighted_dl / n_neutrino_weighted_dl
                # errors on n_neutrino can be neglected, since statistics large
                # gaussian error prop: err_cont = abs(1/n_neutr) * delta n_mupage

                n_mupage_err_dl = np.sqrt(np.count_nonzero(mupage_sel_dl))
                n_rn_err_dl = np.sqrt(np.count_nonzero(rn_sel_dl))
                # n_mupage_weighted_err_dl = n_mupage_weighted_dl * fract_n_mupage_err_to_n_mupage_dl
                # muon_cont_weighted_dl_err.append((np.abs(1/np.count_nonzero(neutr_sel_dl)) * n_mupage_err_dl) * 100)
                muon_cont_weighted_dl_err.append(((n_mupage_err_dl * 33.2137) / n_neutrino_weighted_dl) * 100)
                rn_cont_weighted_dl_err.append(((n_rn_err_dl * 332.87896624) / n_neutrino_weighted_dl) * 100)

            # std

            # if np.count_nonzero(mupage_sel_std) != 0:
            #     fract_n_mupage_err_to_n_mupage_std = n_mupage_err_std / np.count_nonzero(mupage_sel_std)
            # else:
            #     fract_n_mupage_err_to_n_mupage_std = 0

            neutr_sel_std = muon_score_std[is_neutrino_e_cut_std] < cuts[i]
            n_neutrino_weighted_std = np.sum(w_1_y_std[is_neutrino_e_cut_std][neutr_sel_std])

            if n_neutrino_weighted_std != 0:
                mupage_sel_std = muon_score_std[is_mupage_std] < cuts[i]
                rn_sel_std = rn_noise_score_std[is_rn_std] < cuts[i]

                n_mupage_weighted_std = np.sum(w_1_y_std[is_mupage_std][mupage_sel_std])
                n_rn_weighted_std = np.sum(w_1_y_std[is_rn_std][rn_sel_std])

                neutrino_eff_weighted_std.append((n_neutrino_weighted_std / n_neutrinos_total_e_cut_weighted) * 100)
                muon_cont_weighted_std.append((n_mupage_weighted_std / n_neutrino_weighted_std) * 100)
                rn_cont_weighted_std.append((n_rn_weighted_std / n_neutrino_weighted_std) * 100)

                # we calculate n_mupage_weighted_std / n_neutrino_weighted_std
                # errors on n_neutrino can be neglected, since statistics large
                # gaussian error prop: err_cont = abs(1/n_neutr) * delta n_mupage

                n_mupage_err_std = np.sqrt(np.count_nonzero(mupage_sel_std))
                n_rn_err_std = np.sqrt(np.count_nonzero(rn_sel_std))
                # n_mupage_weighted_err_std = n_mupage_weighted_std * fract_n_mupage_err_to_n_mupage_std
                # muon_cont_weighted_std_err.append((np.abs(1/np.count_nonzero(neutr_sel_std)) * n_mupage_err_std) * 100)
                muon_cont_weighted_std_err.append(((n_mupage_err_std * 33.2137) / n_neutrino_weighted_std) * 100)
                rn_cont_weighted_std_err.append(((n_rn_err_std * 332.87896624) / n_neutrino_weighted_std) * 100)

        # remove all 0 entries with n_muons, only needed for dl, since standard reco cant go to 0 muons anyways
        # idx_last_zero = np.amax(np.where(np.array(y_n_muons_dl) == 0))
        # y_n_muons_dl = y_n_muons_dl[idx_last_zero:]
        # y_neutrino_efficiency_dl = y_neutrino_efficiency_dl[idx_last_zero:]
        # x_contamination_mupage_dl = x_contamination_mupage_dl[idx_last_zero:]

        make_plots()



"""
OrcaNet custom loss functions.
"""
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from orcanet.misc import get_register

# fuzz factor for numerical stability
EPS = tf.constant(1e-7, dtype="float32")
# for loading via toml and orcanet custom objects
loss_functions, _register = get_register()


@_register
def lkl_normal_tfp(y_true, y_pred):
    """ Normal distribution using tfp. See lkl_normal. """
    mu_true = y_true[:, 0]
    mu_pred, sigma_pred = y_pred[:, 0], y_pred[:, 1]

    return -1 * tfp.distributions.Normal(
        loc=mu_pred,
        scale=tf.math.maximum(sigma_pred, EPS),
    ).log_prob(mu_true)


@_register
def lkl_normal(y_true, y_pred):
    """
    Negative normal log-likelihood function for n regression output neurons
    with clipping for increased stability.

    For stability in the case of outliers, the loss l_i is capped
    at a maximum of 10 * |pred_i - true_i| for each sample.

    Parameters
    ----------
    y_true : tf.Tensor
        Shape (bs, 2, n) or (bs, 2).
        y_true[:, 0] is the label of shape (bs, n) (true), and y_true[:, 1]
        is not used (necessary as tf 2.1 requires y_true and y_pred to
        have same shape).
    y_pred : tf.Tensor
        Shape (bs, 2, n) or (bs, 2).
        The output of the network.
        y_pred[:, 0] is mu, and y_pred[:, 1] is sigma.

    """
    mu_true = y_true[:, 0]
    mu_pred, sigma_pred = y_pred[:, 0], y_pred[:, 1]

    return _normal_lkl(mu_pred=mu_pred, mu_true=mu_true, sigma_pred=sigma_pred, clip=True)


def _normal_lkl(mu_pred, mu_true, sigma_pred, clip=False, clip_thresh=10):
    delta = mu_pred - mu_true
    std_sq = sigma_pred**2
    loglike = tf.math.log(std_sq + EPS) + delta**2 / (std_sq + EPS)
    if clip:
        loglike = tf.minimum(loglike, clip_thresh*tf.abs(delta))
    return 0.5 * (tf.constant(np.log(2*np.pi), dtype="float32") + loglike)



import matplotlib.pyplot as plt
import argparse
import warnings
from orcanet.history import HistoryHandler
from orcanet.utilities.visualization import TrainValPlotter


class Summarizer:
    """
    Summarize one or more trainings by giving their orcanet folder(s).

    - Plot the training and validation curves in a single plot and show them
    - Print info about the best and worst epochs

    Parameters
    ----------
    folders : str or List, optional
        Path to a orcanet folder, or to multiple folder as a list.
        [default: CWD].
    metric : str or List, optional
        The metric to plot [default: 'loss'].
        If its a list: Same length as folders. Plot a different metric for
        each folder.
    smooth : int, optional
        Apply gaussian blur to the train curve with given sigma.
    labels : List, optional
        Labels for each folder.
    noplot : bool
        Dont plot the train/val curves [default: False].
    width : float
        Scaling of the width of the curves and the marker size [default: 1].

    """
    def __init__(self, folders,
                 metric="loss",
                 smooth=None,
                 labels=None,
                 noplot=False,
                 width=1.,
                 verbose=True):
        if not folders:
            folders = ["./"]
        elif isinstance(folders, str):
            folders = [folders]
        self.folders = folders

        if isinstance(metric, str):
            metric = [metric]
        if len(metric) == 1:
            self.metrics = metric * len(self.folders)
            self._unique_metrics = False
        else:
            if len(metric) != len(folders):
                raise ValueError("Need to give exactly one metric per folder!")
            self.metrics = metric
            self._unique_metrics = True

        if labels is None:
            self.labels = self.folders
        else:
            self.labels = labels

        self.verbose = verbose
        self.smooth = smooth
        self.noplot = noplot
        self.width = width
        self._tvp = None

    def summarize(self, show=True):
        if not self.noplot:
            self._tvp = TrainValPlotter()

        min_stats, max_stats = [], []
        if self.verbose:
            print("Reading stats of {} trainings...".format(len(self.folders)))
        for folder_no in range(len(self.folders)):
            try:
                min_stat, max_stat = self._summarize_folder(folder_no)
                if min_stat is not None:
                    min_stats.append(min_stat)
                if max_stat is not None:
                    max_stats.append(max_stat)
            except Exception as e:
                if self.verbose:
                    print(
                        f"Warning: Can not summarize {self.folders[folder_no]}"
                        f", skipping... ({e})"
                    )

        if self._unique_metrics:
            column_title, y_label = ("combined metrics",) * 2
        else:
            column_title, y_label = self._full_metrics[0], self._metric_names[0]

        if self.verbose:
            if len(min_stats) > 0:
                min_stats.sort()
                print("\nMinimum\n-------")
                print("{}   \t{}\t{}\t{}".format(" ", "Epoch", column_title, "name"))
                for i, stat in enumerate(min_stats, 1):
                    print("{} | \t{}\t{}\t{}".format(i, stat[2], stat[0], stat[1]))

            if len(max_stats) > 0:
                max_stats.sort(reverse=True)
                print("\nMaximum\n-------")
                print("{}   \t{}\t{}\t{}".format(" ", "Epoch", column_title, "name"))
                for i, stat in enumerate(max_stats, 1):
                    print("{} | \t{}\t{}\t{}".format(i, stat[2], stat[0], stat[1]))

        if not self.noplot:
            self._tvp.apply_layout(
                x_label="Epoch",
                y_label=y_label,
                grid=True,
                legend=True,
            )
            if show:
                plt.show()

    @property
    def _metric_names(self):
        """ E.g. [loss, ...] """
        metric_names = []
        for metric in self.metrics:
            if metric.startswith("train_"):
                m = metric[6:]
            elif metric.startswith("val_"):
                m = metric[4:]
            else:
                m = metric
            metric_names.append(m)
        return metric_names

    @property
    def _full_metrics(self):
        """ E.g. [val_loss, ...] """
        full_metrics = []
        for metric in self.metrics:
            if metric.startswith("train_") or metric.startswith("val_"):
                full_metrics.append(metric)
            else:
                full_metrics.append("val_" + metric)
        return full_metrics

    def _summarize_folder(self, folder_no):
        label = self.labels[folder_no]
        folder = self.folders[folder_no]

        hist = HistoryHandler(folder)
        val_data, min_stat, max_stat = None, None, None
        # read data from summary file
        try:
            smry_met_name = self._full_metrics[folder_no]
            max_line = hist.get_best_epoch_info(
                metric=smry_met_name, mini=False)
            min_line = hist.get_best_epoch_info(
                metric=smry_met_name, mini=True)
            min_stat = [min_line[smry_met_name], label, min_line["Epoch"]]
            max_stat = [max_line[smry_met_name], label, max_line["Epoch"]]

            summary_data = hist.get_summary_data()
            val_data = [summary_data["Epoch"],
                        summary_data[self._full_metrics[folder_no]]]
        except OSError:
            if self.verbose:
                print(f"Warning: No summary file found for {folder}")

        except ValueError as e:
            if self.verbose:
                print(f"Error reading summary file {hist.summary_file} ({e})")

        # read data from training files
        full_train_data = hist.get_train_data()
        train_data = [full_train_data["Batch_float"],
                      full_train_data[self._metric_names[folder_no]]]

        if not self.noplot:
            if len(self.labels) == 1:
                train_label, val_label = "training", "validation"
            elif val_data is None:
                train_label, val_label = label, None
            else:
                train_label, val_label = None, label

            self._tvp.plot_curves(train_data=train_data,
                                  val_data=val_data,
                                  train_label=train_label,
                                  val_label=val_label,
                                  smooth_sigma=self.smooth,
                                  tlw=0.5*self.width,
                                  vlw=0.5*self.width,
                                  vms=3*self.width**0.5)
        return min_stat, max_stat

    def summarize_dirs(self):
        """
        Get the best and worst epochs of all given folders as a dict.

        Returns
        -------
        minima : dict
            Keys : Name of folder.
            Values : [Epoch, metric] of where the metric is lowest.
        maxima : dict
            As above, but for where the metric is highest.

        """
        minima, maxima = {}, {}
        for folder_no, folder in enumerate(self.folders):
            hist = HistoryHandler(folder)
            smry_met_name = self._full_metrics[folder_no]
            try:
                max_line = hist.get_best_epoch_info(metric=smry_met_name,
                                                    mini=False)
                min_line = hist.get_best_epoch_info(metric=smry_met_name,
                                                    mini=True)
            except OSError as e:
                warnings.warn(str(e))
                continue

            minima[folder] = [min_line["Epoch"], min_line[smry_met_name]]
            maxima[folder] = [max_line["Epoch"], max_line[smry_met_name]]
        return minima, maxima


def summarize(**kwargs):
    for key in list(kwargs.keys()):
        if kwargs[key] is None:
            kwargs.pop(key)
    Summarizer(**kwargs).summarize()


def get_parser():
    parser = argparse.ArgumentParser(
        description=str(Summarizer.__doc__),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument('folders', type=str, nargs='*')
    parser.add_argument('--metric', type=str, nargs="*")
    parser.add_argument('--smooth', nargs="?", type=int)
    parser.add_argument('--width', nargs="?", type=float)
    parser.add_argument('--labels', nargs="*", type=str)
    parser.add_argument('--noplot', action="store_true")
    return parser


def main():
    parser = get_parser()
    args = vars(parser.parse_args())
    summarize(**args)


if __name__ == '__main__':
    main()



from unittest import TestCase
import numpy as np
import orcanet.lib.label_modifiers as lmods


class TestRegressionLabels(TestCase):
    @staticmethod
    def _get_ys(columns="obs1", data_factor=None, model_output="log_obs",**kwargs):
        data = np.ones((5,))
        if data_factor is not None:
            data *= data_factor
        info_blob = {"y_values": data.astype(
            np.dtype([("obs1", float), ("obs2", float), ("obs3", float)])
        )}
        lmod = lmods.RegressionLabels(
            columns=columns,
            model_output=model_output,
            **kwargs,
        )
        return lmod(info_blob)

    def test_keys(self):
        ys = self._get_ys(log10=True)
        self.assertTrue("log_obs" in ys)

    def test_content_shape1(self):
        ys = self._get_ys(log10=True)
        np.testing.assert_array_equal(ys["log_obs"], np.zeros((5, 1)))

    def test_content_shape1_no_outputname(self):
        ys = self._get_ys(log10=True, model_output=None)
        np.testing.assert_array_equal(ys["obs1"], np.zeros((5, 1)))

    def test_content_stacks_shape1(self):
        ys = self._get_ys(log10=True, stacks=2)
        np.testing.assert_array_equal(ys["log_obs"], np.zeros((5, 2, 1)))

    def test_content(self):
        ys = self._get_ys(columns=["obs1", "obs2", "obs3"])
        np.testing.assert_array_equal(ys["log_obs"], np.ones((5, 3)))

    def test_content_stacks(self):
        ys = self._get_ys(columns=["obs1", "obs2", "obs3"], stacks=2)
        np.testing.assert_array_equal(ys["log_obs"], np.ones((5, 2, 3)))

    def test_log10_invalid_elements_produce_output_zeroes(self):
        ys = self._get_ys(
            columns="obs1",
            data_factor=np.array([100, 10, 1, 0, -1]),
            log10=True,
        )
        np.testing.assert_array_equal(
            ys["log_obs"], np.array([2, 1, 0, 1, 1]).reshape(5, 1))


class TestRegressionLabelsSplit(TestCase):
    def setUp(self):
        data = np.ones((5,))
        self.info_blob = {"y_values": data.astype(
            np.dtype([("obs1", float), ("obs2", float), ("obs3", float)])
        )}
        self.lmod = lmods.RegressionLabelsSplit(
            columns="obs1",
            model_output="log_obs",
        )

    def test_keys(self):
        ys = self.lmod(self.info_blob)
        self.assertListEqual(list(ys.keys()), ["log_obs", "log_obs_err"])

    def test_content_shape1(self):
        target_shapes = {
            "log_obs": (5, 1),
            "log_obs_err": (5, 2, 1),
        }
        ys = self.lmod(self.info_blob)
        for key, shape in target_shapes.items():
            self.assertTupleEqual(shape, ys[key].shape)

    def test_no_stacks(self):
        lmod = lmods.RegressionLabelsSplit(
            columns="obs1",
            model_output="log_obs",
            stacks=2,
        )
        self.assertTrue(lmod.stacks is None)

    def test_yvalues_is_none(self):
        info_blob = {"y_values": None}
        self.assertIsNone(self.lmod(info_blob))

    def test_yvalues_does_not_have_right_column(self):
        info_blob = {"y_values": np.ones((5,)).astype(
            np.dtype([("asdasd", float), ])
        )}
        self.assertIsNone(self.lmod(info_blob))


class TestClassificationLabels(TestCase):
    @staticmethod
    def _get_ys(
        column="obs1", model_output="cat", classes=None,
    ):
        if classes is None:
            classes = {"class1": [1], "class2": [2]}
        data = np.array([1, 1, 2, 2])
        info_blob = {
            "y_values": data.astype(
                np.dtype([("obs1", float), ("obs2", float), ("obs3", float)])
            )
        }
        lmod = lmods.ClassificationLabels(
            column=column,
            model_output=model_output,
            classes=classes,
        )
        return lmod(info_blob)

    def test_keys(self):
        ys = self._get_ys()
        self.assertTrue("cat" in ys)

    def test_content_shape(self):
        ys = self._get_ys()
        np.testing.assert_array_equal(
            ys["cat"], np.array([[1.0, 0.0], [1.0, 0.0], [0.0, 1.0], [0.0, 1.0]])
        )

    def test_wrong_classes(self):
        with self.assertRaises(KeyError):
            self._get_ys(classes={"class42": [1], "class2": [2]})


class TestTSClassifier(TestCase):
    @staticmethod
    def _get_ys():
        # take 3 events, one muon CC and one muon NC, and one atm muon
        y_values = np.array(
            [(14, 2), (14, 3), (13, None)],
            dtype=[("particle_type", float), ("is_cc", float)],
        )

        info_blob = {}
        info_blob["y_values"] = y_values

        lmod = lmods.TSClassifier(is_cc_convention=2, model_output="TS_output")
        return lmod(info_blob)

    def test_keys(self):
        ys = self._get_ys()
        self.assertTrue("TS_output" in ys)

    def test_content_shape(self):
        ys = self._get_ys()
        np.testing.assert_array_equal(
            ys["TS_output"], np.array([[1.0, 0.0], [0.0, 1.0], [1.0, 0.0]])
        )



import numpy as np
import tensorflow as tf
import orcanet.lib.losses as on_losses


class TestLklNormal(tf.test.TestCase):
    def setUp(self):
        self.loss_func = on_losses.lkl_normal
        self.y_pred = tf.constant([
            [[0], [1]],
            [[1], [1]],
            [[2], [2]],
        ], dtype="float32")  # shape (3, 2, 1)
        self.y_true = tf.constant([
            [[0], [0]],
            [[0], [0]],
            [[0], [0]],
        ], dtype="float32")

        self.target = tf.constant([[0.9189385], [1.4189385], [2.1120858]])

    def test_normal(self):
        loss = self.loss_func(self.y_true, self.y_pred)
        self.assertAllClose(loss, self.target)

    def test_normal_flat(self):
        loss = self.loss_func(tf.squeeze(self.y_true), tf.squeeze(self.y_pred))
        self.assertAllClose(loss, tf.squeeze(self.target))

    def test_normal_double(self):
        y_pred = tf.repeat(self.y_pred, 4, -1)
        y_true = tf.repeat(self.y_true, 4, -1)
        loss = self.loss_func(y_true, y_pred)
        self.assertAllClose(loss, tf.repeat(self.target, 4, -1))

    def test_if_training_works(self):
        inp = tf.keras.Input((3,))
        x = tf.keras.layers.Dense(6)(inp)
        x = tf.keras.layers.Reshape((2, 3))(x)
        model = tf.keras.Model(inp, x)
        model.compile("sgd", loss=self.loss_func)

        xs = np.ones((5, 3))
        ys = np.ones((5, 2, 3))
        model.fit(xs, ys)


class TestLklNormalTfp(TestLklNormal):
    def setUp(self):
        super().setUp()
        self.loss_func = on_losses.lkl_normal_tfp



import h5py
import time
import numpy as np
import tensorflow as tf
import tensorflow.keras as ks


class Hdf5BatchGenerator(ks.utils.Sequence):
    def __init__(self, files_dict,
                 batchsize=64,
                 key_x_values="x",
                 key_y_values="y",
                 sample_modifier=None,
                 label_modifier=None,
                 fixed_batchsize=False,
                 y_field_names=None,
                 phase="training",
                 xs_mean=None,
                 f_size=None,
                 keras_mode=True,
                 shuffle=False,
                 class_weights=None):
        """
        Yields batches of input data from h5 files.

        This will go through one file, or multiple files in parallel, and yield
        one batch of data, which can then be used as an input to a model.
        Since multiple filepaths can be given to read out in parallel,
        this can also be used for models with multiple inputs.

        Parameters
        ----------
        files_dict : dict
            Pathes of the files to train on.
            Keys: The name of every input (from the toml list file, can be multiple).
            Values: The filepath of a single h5py file to read data from.
        batchsize : int
            Batchsize that will be used for reading data from the files.
        key_x_values : str
            The name of the datagroup in the h5 input files which contains
            the samples for the network.
        key_y_values : str
            The name of the datagroup in the h5 input files which contains
            the info for the labels. If this name is not in the file,
            y_values will be set to None.
        sample_modifier : function or None
            Operation to be performed on batches of samples read from the input
            files before they are fed into the model.
        y_field_names : tuple or list or str, optional
            During train and val, read out only these fields from the y dataset.
            --> Speed up, especially if there are many fields.
        phase : str
            Which phase are we in? training, validation, or inference.
            Inference means both orga.predict and orga.inference, i.e.
            whenever we write a h5 file.
        label_modifier : function or None
            Operation to be performed on batches of labels read from the input files
            before they are fed into the model.
        fixed_batchsize : bool
            The last batch in the file might be smaller then the batchsize.
            Usually, this is no problem, but set to True to pad this batch to
            given batchsize.
        xs_mean : ndarray or None
            Zero center image to be subtracted from data as preprocessing.
        f_size : int or None
            Specifies the number of samples to be read from the .h5 file.
            If none, the whole .h5 file will be used.
        keras_mode : bool
            If true, yield xs and ys (samples and labels) for the keras fit
            generator function.
            If false, yield the info_blob containing the full sample and label
            info, both before and after the modifiers have been applied.
        shuffle : bool
            Randomize the order in which batches are read from the file
            (once during init). Can reduce read out speed.

        """
        if phase not in ("training", "validation", "inference"):
            raise ValueError("Invalid phase")
        self.files_dict = files_dict
        self.batchsize = batchsize
        self.key_x_values = key_x_values
        self.key_y_values = key_y_values
        self.sample_modifier = sample_modifier
        self.label_modifier = label_modifier
        self.fixed_batchsize = fixed_batchsize
        self.phase = phase
        self.xs_mean = xs_mean
        self.f_size = f_size
        self.keras_mode = keras_mode
        self.shuffle = shuffle
        self.class_weights = class_weights

        if y_field_names is not None:
            if isinstance(y_field_names, str):
                y_field_names = (y_field_names, )
            else:
                y_field_names = tuple(y_field_names)
        self.y_field_names = y_field_names

        # a dict with the names of list inputs as keys, and the opened
        # h5 files as values
        self._files = {}
        # start index of each batch in the file
        self._sample_pos = None
        # total number of samples per file
        self._total_f_size = None

        # for keeping track of the readout speed
        self._total_time = 0.
        self._total_batches = 0
        self._file_meta = None

        self.open()

    def __len__(self):
        """ Number of batches in the Sequence (includes queue). """
        return len(self._sample_pos)

    def __getitem__(self, index):
        """
        Gets batch number `index`.

        Returns
        -------
        xs : dict
            Samples for the model train on.
            Keys : str
                The name(s) of the input layer(s) of the model.
            Values : ndarray or tuple
                A batch of samples for the corresponding input.
                If x is an indexed datasets, this will be a tuple instead,
                with [0] being the values, and [1] being the number of
                items per sample.
        ys : dict or None
            Labels for the model to train on. Will be None if there are
            no labels in the file.
            Keys : str
                The name(s) of the output layer(s) of the model.
            Values : ndarray
                A batch of labels for the corresponding output.

        If class_weights is not None, will return aditionally:
        sample_weights : dict
            Maps output names to weights for each sample in the batch as a
            np.array.

        If keras_mode is False, will return instead:
        info_blob : dict
            Blob containing the x_values, y_values, xs and ys, and optionally
            the sample_weights.

        """
        start_time = time.time()
        file_index = self._sample_pos[index]
        info_blob = {"phase": self.phase, "meta": self.get_file_meta()}
        info_blob["x_values"] = self.get_x_values(file_index)
        info_blob["y_values"] = self.get_y_values(file_index)

        # Modify the samples
        if self.sample_modifier is not None:
            xs = self.sample_modifier(info_blob)
        else:
            xs = info_blob["x_values"]
        info_blob["xs"] = xs

        # Modify the labels
        if info_blob["y_values"] is not None and self.label_modifier is not None:
            ys = self.label_modifier(info_blob)
        else:
            ys = None
        info_blob["ys"] = ys

        if self.fixed_batchsize:
            self.pad_to_size(info_blob)

        if self.class_weights is not None:
            info_blob["sample_weights"] = _get_sample_weights(ys, self.class_weights)

        self._total_time += time.time() - start_time
        self._total_batches += 1
        if self.keras_mode:
            if info_blob.get("sample_weights"):
                return info_blob["xs"], info_blob["ys"], info_blob["sample_weights"]
            else:
                return info_blob["xs"], info_blob["ys"]
        else:
            return info_blob

    def pad_to_size(self, info_blob):
        """ Pad the batch to have a fixed batchsize. """
        org_batchsize = next(iter(info_blob["xs"].values())).shape[0]
        if org_batchsize == self.batchsize:
            return
        info_blob["org_batchsize"] = org_batchsize
        for input_key, x in info_blob["xs"].items():
            info_blob["xs"][input_key] = _pad_to_size(x, self.batchsize)
        if info_blob.get("ys") is not None:
            for output_key, y in info_blob["ys"].items():
                info_blob["ys"][output_key] = _pad_to_size(y, self.batchsize)

    def open(self):
        """ Open all files and prepare for read out. """
        for input_key, file in self.files_dict.items():
            self._files[input_key] = h5py.File(file, 'r')
        self._store_file_length()
        self._store_batch_indices()

    def close(self):
        """ Close all files again. """
        for f in list(self._files.values()):
            f.close()

    def get_x_values(self, start_index):
        """
        Read one batch of samples from the files and zero center.

        Parameters
        ----------
        start_index : int
            The start index in the h5 files at which the batch will be read.
            The end index will be the start index + the batch size.

        Returns
        -------
        x_values : dict
            One batch of data for each input file.

        """
        x_values = {}
        for input_key, file in self._files.items():
            slc = slice(start_index, start_index + self._batchsize)

            ix_dset_name = _get_indexed_dset_name(file, self.key_x_values)
            if ix_dset_name is None:
                # normal dataset
                x_values[input_key] = file[self.key_x_values][slc]
            else:
                # indexed dataset: adjust slice according to indices
                indices = file[ix_dset_name][slc]
                slc = slice(
                    indices[0]["index"],
                    indices[-1]["index"] + indices[-1]["n_items"],
                )
                x_values[input_key] = (file[self.key_x_values][slc], indices["n_items"])

            if self.xs_mean is not None:
                x_values[input_key] = np.subtract(
                    x_values[input_key], self.xs_mean[input_key])

        return x_values

    def get_y_values(self, start_index):
        """
        Get y_values for the nn. Since the y_values are hopefully the same
        for all the files, use the ones from the first. TODO add check

        Parameters
        ----------
        start_index : int
            The start index in the h5 files at which the batch will be read.
            The end index will be the start index + the batch size.

        Returns
        -------
        y_values : ndarray
            The y_values, right from the files.

        """
        first_file = list(self._files.values())[0]
        try:
            slc = slice(start_index, start_index + self._batchsize)
            if self.y_field_names is not None and self.phase != "inference":
                y_values = first_file[self.key_y_values][(slc,) + tuple(self.y_field_names,)]
                if len(self.y_field_names) == 1:
                    # result of slice is a ndarray; convert to structured
                    y_values = y_values.astype(
                        np.dtype([(self.y_field_names[0], y_values.dtype)]))
            else:
                y_values = first_file[self.key_y_values][slc]
        except KeyError:
            # can not look up y_values, lets hope we dont need them
            y_values = None
        return y_values

    def print_timestats(self, print_func=None):
        """ Print stats about how long it took to read batches. """
        if print_func is None:
            print_func = print
        print_func("Statistics of data readout:")
        print_func(f"\tTotal time:\t{self._total_time/60:.2f} min")
        if self._total_batches != 0:
            print_func(
                f"\tPer batch:\t"
                f"{1000 * self._total_time/self._total_batches:.5} ms"
            )

    def get_file_meta(self):
        """ Meta information about the files. Only read out once. """
        if self._file_meta is None:
            self._file_meta = {}
            # sample and label dataset for each input
            datasets = {}
            for input_key, file in self._files.items():
                datasets[input_key] = {
                    "samples": file[self.key_x_values],
                    "samples_is_indexed": _get_indexed_dset_name(
                        file, self.key_x_values) is not None,
                    "labels": file[self.key_y_values],
                }
            self._file_meta["datasets"] = datasets
        return self._file_meta

    @property
    def _size(self):
        """ Size of the files that will be read in. Can be smaller than the actual
        file size if defined by user. """
        if self.f_size is None:
            return self._total_f_size
        else:
            return self.f_size

    @property
    def _batchsize(self):
        """
        Return the effective batchsize. Can be smaller than the user defined
        one if it would be larger than the size of the file.
        """
        if self._size < self.batchsize:
            return self._size
        else:
            return self.batchsize

    def _store_file_length(self):
        """
        Make sure all files have the same length and store this length.
        """
        lengths = []
        for f in list(self._files.values()):
            ix_dset_name = _get_indexed_dset_name(f, self.key_x_values)
            if ix_dset_name is None:
                dset_name = self.key_x_values
            else:
                dset_name = ix_dset_name
            lengths.append(len(f[dset_name]))

        if not lengths.count(lengths[0]) == len(lengths):
            self.close()
            raise ValueError("All data files must have the same length! "
                             "Given were:\n " + str(lengths))

        self._total_f_size = lengths[0]

    def _store_batch_indices(self):
        """
        Define the start indices of each batch in the h5 file and store this.
        """
        if self.phase == "inference":
            # for inference: take all batches
            total_no_of_batches = np.ceil(self._size / self._batchsize)
        else:
            # else: skip last batch if it has too few event for a full batch
            # this is mostly because tf datasets can't be used
            # with variable batchsize (status tf 2.5)
            total_no_of_batches = np.floor(self._size / self._batchsize)

        sample_pos = np.arange(int(total_no_of_batches)) * self._batchsize
        if self.shuffle:
            np.random.shuffle(sample_pos)

        self._sample_pos = sample_pos


def _get_indexed_dset_name(file, dset):
    """ If this is an indexed dataset, return the name of the indexed set. """
    dset_name_indexed = f"{dset}_indices"
    if file[dset].attrs.get("indexed") and dset_name_indexed in file:
        return dset_name_indexed
    else:
        return None


def _get_sample_weights(ys, class_weights):
    """
    Produce a weight for each sample given the weight for each class.

    Parameters
    ----------
    ys : dict
        Maps output names to categorical one-hot labels as np.arrays.
        Expected to be 2D (n_samples, n_classes).
    class_weights : dict
        Maps output neuron numbers to weights as floats.

    Returns
    -------
    sample_weights : dict
        Maps output names to weights for each sample in the batch as a
        np.array.

    """
    sample_weights = {}
    for output_name, labels in ys.items():
        class_weights_arr = np.ones(labels.shape[1])
        for k, v in class_weights.items():
            class_weights_arr[int(k)] = v
        labels_class = np.argmax(labels, axis=-1)
        sample_weights[output_name] = class_weights_arr[labels_class]
    return sample_weights


def get_h5_generator(orga, files_dict, f_size=None, zero_center=False,
                     keras_mode=True, shuffle=False, use_def_label=True,
                     phase="training"):
    """
    Initialize the hdf5_batch_generator_base with the paramters in orga.cfg.

    Parameters
    ----------
    orga : orcanet.core.Organizer
        Contains all the configurable options in the OrcaNet scripts.
    files_dict : dict
        Pathes of the files to train on.
        Keys: The name of every input (from the toml list file, can be multiple).
        Values: The filepath of a single h5py file to read samples from.
    f_size : int or None
        Specifies the number of samples to be read from the .h5 file.
        If none, the whole .h5 file will be used.
    zero_center : bool
        Whether to use zero centering.
        Requires orga.zero_center_folder to be set.
    keras_mode : bool
        Specifies if mc-infos (y_values) should be yielded as well. The
        mc-infos are used for evaluation after training and testing is finished.
    shuffle : bool
        Randomize the order in which batches are read from the file.
        Significantly reduces read out speed.
    use_def_label : bool
        If True and no label modifier is given by user, use the default
        label modifier instead of none.

    Yields
    ------
    xs : dict
        Data for the model train on.
            Keys : str  The name(s) of the input layer(s) of the model.
            Values : ndarray    A batch of samples for the corresponding input.
    ys : dict or None
        Labels for the model to train on.
            Keys : str  The name(s) of the output layer(s) of the model.
            Values : ndarray    A batch of labels for the corresponding output.
        Will be None if there are no labels in the file.
    y_values : ndarray, optional
        Y values from the file. Only yielded if yield_mc_info is True.

    """
    if orga.cfg.label_modifier is not None:
        label_modifier = orga.cfg.label_modifier
    elif use_def_label:
        assert orga._auto_label_modifier is not None, \
            "Auto label modifier has not been set up"
        label_modifier = orga._auto_label_modifier
    else:
        label_modifier = None

    # get xs_mean or load/create if not stored yet
    if zero_center:
        xs_mean = orga.get_xs_mean()
    else:
        xs_mean = None

    generator = Hdf5BatchGenerator(
        files_dict=files_dict,
        batchsize=orga.cfg.batchsize,
        key_x_values=orga.cfg.key_x_values,
        key_y_values=orga.cfg.key_y_values,
        sample_modifier=orga.cfg.sample_modifier,
        label_modifier=label_modifier,
        phase=phase,
        xs_mean=xs_mean,
        f_size=f_size,
        keras_mode=keras_mode,
        shuffle=shuffle,
        class_weights=orga.cfg.class_weight,
        fixed_batchsize=orga.cfg.fixed_batchsize,
        y_field_names=orga.cfg.y_field_names,
    )

    return generator


def make_dataset(gen):
    output_signature = tuple([{k: _get_spec(v) for k, v in d.items()} for d in gen[0]])
    return tf.data.Dataset.from_generator(
        lambda: gen, output_signature=output_signature)


def _get_spec(x):
    if isinstance(x, tf.RaggedTensor):
        return tf.RaggedTensorSpec.from_value(x)
    else:
        return tf.TensorSpec(
            shape=x.shape,
            dtype=x.dtype,
        )


def _pad_to_size(x, size):
    """ Pad x to given size along axis 0 by repeating last element. """
    length = x.shape[0]
    if length > size:
        raise ValueError(f"Can't pad x with shape {x.shape} to length {size}")
    elif length == size:
        return x
    else:
        if tf.is_tensor(x):
            f_conc = tf.concat
        else:
            f_conc = np.concatenate

        return f_conc([x] + [x[-1:]] * (size - length), axis=0)



"""
Use orga.predict with a parser.

Usage:
    parser_orcapred.py FOLDER LIST CONFIG MODEL [--epoch EPOCH] [--fileno FILENO]
    parser_orcapred.py (-h | --help)

Arguments:
    FOLDER  Path to the folder where everything gets saved to, e.g. the
            summary.txt, the plots, the trained models, etc.
    LIST    A .toml file which contains the pathes of the training and
            validation files. An example can be found in
            examples/list_file.toml
    CONFIG  A .toml file which sets up the training. An example can be
            found in examples/config_file.toml. The possible parameters
            are listed in core.py in the class Configuration.
    MODEL   Path to a .toml file with infos about a model.
            An example can be found in examples/explanation.toml.

Options:
    -h --help        Show this screen.
    --epoch=EPOCH    Use model of given epoch. [default: None]
    --fileno=FILENO  Use model of given fileno. [default: None]

"""
import warnings
from matplotlib import use
use('Agg')

from docopt import docopt

from orcanet.core import Organizer
from orcanet_contrib.orca_handler_util import update_objects


def orca_pred(output_folder, list_file, config_file, model_file,
              epoch=None, fileno=None):
    """
    Run orga.predict with predefined ModelBuilder networks using a parser.

    Per default, the most recent saved model will be loaded.

    Parameters
    ----------
    output_folder : str
        Path to the folder where everything gets saved to, e.g. the summary
        log file, the plots, the trained models, etc.
    list_file : str
        Path to a list file which contains pathes to all the h5 files that
        should be used for training and validation.
    config_file : str
        Path to a .toml file which overwrites some of the default settings
        for training and validating a model.
    model_file : str
        Path to a file with parameters to build a model of a predefined
        architecture with OrcaNet.
    epoch : int, optional
        The epoch of the saved model to predict with.
    fileno : int, optional
        The filenumber of the saved model to predict with.

    """
    # Set up the Organizer with the input data
    orga = Organizer(output_folder, list_file, config_file, tf_log_level=1)

    # When predicting with a orga model, the right modifiers and custom
    # objects need to be given
    update_objects(orga, model_file)

    # Per default, a prediction will be done for the model with the
    # highest epoch and filenumber.
    orga.predict(epoch=epoch, fileno=fileno)


def main():
    """ Run the orca_pred function with a parser. """
    warnings.warn("parser_orcapred is deprecated! Use orcanet predict instead.")
    args = docopt(__doc__)
    if args["--epoch"] == "None":
        epoch = None
    else:
        epoch = int(args["--epoch"])

    if args["--fileno"] == "None":
        fileno = None
    else:
        fileno = int(args["--fileno"])

    orca_pred(output_folder=args['FOLDER'],
              list_file=args['LIST'],
              config_file=args['CONFIG'],
              model_file=args['MODEL'],
              epoch=epoch,
              fileno=fileno)


if __name__ == '__main__':
    main()



from unittest import TestCase
import numpy as np
import orcanet.lib.sample_modifiers as smods


class TestAppli(TestCase):
    def setUp(self):
        self.inp_a_name = "inp_A"
        self.batchsize = 32
        self.info_blob = {
            "x_values": {
                self.inp_a_name: np.ones((self.batchsize, 5, 7, 11))
            }
        }

    def test_reshape(self):
        new_shape = (5*7, 11)
        result = smods.Reshape(new_shape)(self.info_blob)
        self.assertEqual(
            result[self.inp_a_name].shape,
            (self.batchsize, ) + new_shape
        )

    def test_permute(self):
        result = smods.Permute((3, 2, 1))(self.info_blob)
        self.assertEqual(
            result[self.inp_a_name].shape,
            (self.batchsize, ) + (11, 7, 5)
        )

    def test_joined_modifier(self):
        new_shape = (5*7, 11)
        modifiers = [
            smods.Reshape(new_shape),
            smods.Permute((2, 1))
        ]

        result = smods.JoinedModifier(modifiers)(self.info_blob)
        self.assertEqual(
            result[self.inp_a_name].shape,
            (self.batchsize, ) + tuple(reversed(new_shape))
        )


class TestGraphSampleMod(TestCase):
    def setUp(self) -> None:
        self.smod = smods.GraphEdgeConv(
            column_names=("c1", "time"),
            node_features=("c1", ),
            coord_features=("c1", "time"),
            knn=10,
        )
        self.info_blob = {
            "x_values": {"file_1":
                (
                    # hits
                    np.arange(40, dtype="float32").reshape(20, 2),
                    # n_items
                    np.array([5, 15], dtype="float32"),
                )
            },
            "meta": {"datasets": {"file_1": {
                "samples_is_indexed": True,
            }}}
        }

    def test_nodes_shape(self):
        out = self.smod(self.info_blob)
        self.assertEqual(
            list(out["nodes"].shape),
            [2, None, 1],
        )

    def test_nodes_n_items(self):
        out = self.smod(self.info_blob)
        np.testing.assert_array_equal(
            out["nodes"].row_lengths().numpy(),
            np.array([11, 15], dtype="int32"),
        )

    def test_nodes_content(self):
        out = self.smod(self.info_blob)
        np.testing.assert_array_equal(
            out["nodes"].merge_dims(0, 1).numpy(),
            np.concatenate(
                [np.arange(0, 10, 2), np.zeros(6), np.arange(10, 40, 2)]
            ).astype("float32")[:, None],
        )



import warnings
import numpy as np
import orcanet.misc as misc

# for loading via toml
lmods, register = misc.get_register()


class ColumnLabels:
    """
    Label of each model output is column with the same name in the h5 file.
    This is the default label modifier.

    Example
    -------
    Model has output "energy" --> label is column "energy" from the label
    dataset in the h5 file.

    Parameters
    ----------
    model : ks.Model
            A keras model.

    """

    def __init__(self, model):
        self.output_names = model.output_names

    def __call__(self, info_blob):
        ys = {name: info_blob["y_values"][name] for name in self.output_names}
        return ys


@register
class RegressionLabels:
    """
    Generate labels for regression.

    Parameters
    ----------
    columns : str or list
            Name(s) of the columns in the label dataset that contain the labels.
    model_output : str, optional
            Name of the output of the network.
            Default: Same as columns (only valid if columns is a str).
    log10 : bool
            Take log10 of the labels. Invalid values in the label will produce 0
            and a warning.
    stacks : int, optional
            Stack copies of the label this many times along a new axis at position 1.
            E.g. if the label is shape (?, 3), it will become
            shape (?, stacks, 3). Used for lkl regression.

    Examples
    --------
    >>> RegressionLabels(columns=['dir_x', 'dir_y', 'dir_z'], model_output='dir')
    or in the config.toml:
    label_modifier = {name='RegressionLabels', columns=['dir_x','dir_y','dir_z'], model_output='dir'}
    Will produce array of shape (bs, 3) for model output 'dir'.
    >>> RegressionLabels(columns='dir_x')
    Will produce array of shape (bs, 1) for model output 'dir_x'.

    """

    def __init__(self, columns, model_output=None, log10=False, stacks=None):
        if isinstance(columns, str):
            columns = [
                columns,
            ]
        else:
            columns = list(columns)
        if model_output is None:
            if len(columns) != 1:
                raise ValueError(
                    f"If model_output is not given, columns must be length 1!"
                )
            model_output = columns[0]

        self.columns = columns
        self.model_output = model_output
        self.stacks = stacks
        self.log10 = log10
        self._warned = False

    def __call__(self, info_blob):
        y_values = info_blob["y_values"]
        if y_values is None:
            if not self._warned:
                warnings.warn(f"Can not generate labels: No y_values available!")
                self._warned = True
            return None
        try:
            y_value = y_values[self.columns]
        except KeyError:
            if not self._warned:
                warnings.warn(
                    f"Can not generate labels: {self.columns} " f"not found in y_values"
                )
                self._warned = True
            return None
        y_value = misc.to_ndarray(y_value, dtype="float32")
        return {self.model_output: self.process_label(y_value)}

    def process_label(self, y_value):
        ys = y_value
        if self.log10:
            gr_zero = ys > 0
            if not np.all(gr_zero):
                warnings.warn(
                    "invalid value encountered in log10, setting result to 1",
                    category=RuntimeWarning,
                )
            ys = np.log10(ys, where=gr_zero, out=np.ones_like(ys, dtype="float32"))
        if self.stacks:
            ys = np.repeat(ys[:, None], repeats=self.stacks, axis=1)

        return ys


@register
class RegressionLabelsSplit(RegressionLabels):
    """
    Generate labels for regression.

    Intended for networks that output recos and errs in seperate towers
    (for example when using OutputRegNormalSplit as output layer block).

    Example
    -------
    >>> RegressionLabelsSplit(columns=['dir_x', 'dir_y', 'dir_z'], model_output='dir')
    Will produce label 'dir' of shape (bs, 3),
    and label 'dir_err' of shape (bs, 2, 3).

    'dir_err' is just the label twice, along a new axis at -2.
    Necessary because pred and truth must be the same shape.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.err_output_format = "{}_err"
        if self.stacks is not None:
            warnings.warn(
                "Can not use stacks option with RegressionLabelsSplit, ignoring..."
            )
            self.stacks = None
        self._warned = False

    def __call__(self, info_blob):
        output_dict = super().__call__(info_blob)
        if output_dict is None:
            return None
        err_outputs = {}
        for name, label in output_dict.items():
            err_outputs[self.err_output_format.format(name)] = np.repeat(
                np.expand_dims(label, axis=-2), repeats=2, axis=-2
            )
        output_dict.update(err_outputs)
        return output_dict


@register
class ClassificationLabels:
    """
    One-hot encoding for general purpose classification labels based on one mc label column.

    Parameters
    ----------
    column : str
            Identifier of which mc info to create the labels from.
    classes : dict
            Specify for each class the conditions the column name has to fulfil.
            The keys have to be named "class1", "class2", etc
    model_output : str, optional
            The name of the output layer's outputs.

    Example
    -------
    2-class cf for signal and background; put this into the config.toml:
    label_modifier = {name="ClassificationLabels", column="particle_type", classes={class1 = [12, -12, 14, -14], class2 = [13, -13, 0]}, model_output="bg_output"}

    """

    def __init__(
        self,
        column,
        classes,
        model_output=None,
    ):
        self.column = column
        self.classes = classes
        self.model_output = model_output
        self._warned = False

        if "class1" not in self.classes:
            raise KeyError("Class names must be named 'class1', 'class2',...")
        if not len(self.classes["class1"]) > 0:
            raise ValueError("Not a valid list for a class")

        if model_output is None:
            self.model_output = column

    def __call__(self, info_blob):

        y_values = info_blob["y_values"]

        if y_values is None:
            if not self._warned:
                warnings.warn(f"Can not generate labels: No y_values available!")
                self._warned = True
            return None

        try:
            y_value = y_values[self.column]
        except ValueError:
            if not self._warned:
                warnings.warn(
                    f"Can not generate labels: {self.column} " f"not found in y_values"
                )
                self._warned = True
            #let this pass by for real data
            return None

        # create an array of the final shape, initialized with zeros
        n_classes = len(self.classes)
        batchsize = y_values.shape[0]
        categories = np.zeros((batchsize, n_classes), dtype="bool")

        # iterate over every class and set entries to 1 if condition is fulfilled
        for i in range(n_classes):
            categories[:, i] = np.in1d(
                y_values[self.column], self.classes["class" + str(i + 1)]
            )

        return {self.model_output: categories.astype(np.float32)}


@register
class TSClassifier:

    """
    One-hot encoding for track/shower classifier. Muon neutrino CC are tracks, the rest
    of neutrinos is shower. This means, this has to be extended for tau neutrinos. Atm.
    muon events, if any, are tracks.

    Parameters
    ----------
    is_cc_convention : int
            The convention used in the MC prod to indicate a charged current interaction.
            For post 2020 productions this is 2.
    model_output : str, optional
            Name of the output of the network.
            Default: Same as names (only valid if names is a str).

    Example
    -------
    label_modifier = {name='TSClassifier', is_cc_convention=2}

    """

    def __init__(
        self,
        is_cc_convention,
        model_output="ts_output",
    ):
        self.is_cc_convention = is_cc_convention
        self.model_output = model_output

    def __call__(self, info_blob):

        y_values = info_blob["y_values"]

        try:
            particle_type = y_values["particle_type"]
            is_cc = y_values["is_cc"] == self.is_cc_convention
        except ValueError:
            if not self._warned:
                warnings.warn(
                    f"Can not generate labels: particle_type or is_cc not found in y_values"
                )
                self._warned = True
            #let this pass by for real data
            return None

        ys = dict()

        # create conditions from particle_type and is cc
        is_muon_cc = np.logical_and(np.abs(particle_type) == 14, is_cc)

        # in case there are atm. muon events in the mix as well, declare them to be tracks
        is_track = np.logical_or(is_muon_cc, np.abs(particle_type) == 13)

        is_shower = np.invert(is_track)

        batchsize = y_values.shape[0]
        # categorical [shower, track] -> [1,0] = shower, [0,1] = track
        categorical_ts = np.zeros((batchsize, 2), dtype="bool")

        categorical_ts[:, 0] = is_track
        categorical_ts[:, 1] = is_shower

        ys[self.model_output] = categorical_ts.astype(np.float32)

        return ys



from unittest import TestCase
import os
import numpy as np
from orcanet.history import HistoryHandler


class TestHistoryHandler(TestCase):
    """
    Test the in out Handler on 2 dummy summary files
    (summary.txt, summary_2.txt).
    """
    def setUp(self):
        self.output_folder = os.path.join(os.path.dirname(__file__),
                                          "data", "dummy_model")
        self.history = HistoryHandler(self.output_folder)

        self.summary_filename_2 = "summary_2.txt"

    def test_get_metrics(self):
        metrics = self.history.get_metrics()
        target = ["loss", "acc"]
        np.testing.assert_array_equal(metrics, target)

    def test_get_summary_data(self):
        summary_data = self.history.get_summary_data()
        target = np.array(
            [(0.03488, 0.005, 0.07776, 0.05562, 0.971, 0.9808),
             (0.06971, 0.00465, 0.05034, np.nan, 0.9822, np.nan)],
            dtype=[('Epoch', '<f8'), ('LR', '<f8'), ('train_loss', '<f8'),
                   ('val_loss', '<f8'), ('train_acc', '<f8'), ('val_acc', '<f8')])
        assert_equal_struc_array(summary_data, target)

    def test_get_train_data(self):
        train_data = self.history.get_train_data()
        print(train_data)
        target = np.array(
            [(250., 0.00018755, 0.487258, 0.756687),
             (500., 0.00056265, 0.291661, 0.874188),
             (250., 0.0350691, 0.0589843, 0.977938),
             (500., 0.0354442, 0.0554589, 0.980375)],
            dtype=[('Batch', '<f8'), ('Batch_float', '<f8'),
                   ('loss', '<f8'), ('acc', '<f8')])
        assert_equal_struc_array(train_data, target)

    def test_get_column_names(self):
        column_names = self.history.get_column_names()
        target = ('Epoch', 'LR', 'train_loss', 'val_loss', 'train_acc', 'val_acc')
        self.assertSequenceEqual(column_names, target)

    def test_get_state(self):
        self.history.summary_filename = self.summary_filename_2

        state = self.history.get_state()
        print(state)
        target = [
            {'epoch': 0.1, 'is_trained': True, 'is_validated': True},
            {'epoch': 0.2, 'is_trained': False, 'is_validated': True},
            {'epoch': 0.3, 'is_trained': True, 'is_validated': False},
            {'epoch': 0.4, 'is_trained': False, 'is_validated': False},
        ]
        self.assertSequenceEqual(state, target)

    def test_plot_metric_unknown_metric(self):
        with self.assertRaises(ValueError):
            self.history.plot_metric("test")
    """
    @patch('orcanet.history.plot_history')
    def test_plot_metric_loss(self, mock_plot_history):
        def plot_history(train_data, val_data, **kwargs):
            return train_data, val_data, kwargs
        mock_plot_history.side_effect = plot_history

        value_train, value_val, value_kwargs = self.history.plot_metric("loss")

        target_train = [
            np.array([0.000187551, 0.000562653, 0.0350691, 0.0354442]),
            np.array([0.487258, 0.291661, 0.0589843, 0.0554589]),
        ]

        target_val = [
            np.array([0.03488, 0.06971]),
            np.array([0.05562, np.nan]),
        ]
        target_kwargs = {'y_label': 'loss'}

        np.testing.assert_array_almost_equal(target_train, value_train)
        np.testing.assert_array_almost_equal(target_val, value_val)
        self.assertDictEqual(target_kwargs, value_kwargs)

    @patch('orcanet.history.plot_history')
    def test_plot_metric_acc(self, mock_plot_history):
        def plot_history(train_data, val_data, **kwargs):
            return train_data, val_data, kwargs
        mock_plot_history.side_effect = plot_history

        value_train, value_val, value_kwargs = self.history.plot_metric("acc")

        target_train = [
            np.array([0.000187551, 0.000562653, 0.0350691, 0.0354442]),
            np.array([0.756687, 0.874188, 0.977938, 0.980375]),
        ]

        target_val = [
            np.array([0.03488, 0.06971]),
            np.array([0.9808, np.nan]),
        ]
        target_kwargs = {'y_label': 'acc'}

        np.testing.assert_array_almost_equal(target_train, value_train)
        np.testing.assert_array_almost_equal(target_val, value_val)
        self.assertDictEqual(target_kwargs, value_kwargs)

    @patch('orcanet.history.plot_history')
    def test_plot_lr(self, mock_plot_history):
        def plot_history(train_data, val_data, **kwargs):
            return train_data, val_data, kwargs
        mock_plot_history.side_effect = plot_history

        value_train, value_val, value_kwargs = self.history.plot_lr()

        target_train = None

        target_val = [
            np.array([0.03488, 0.06971]),
            np.array([0.005, 0.00465])
        ]
        target_kwargs = {
            'y_label': 'Learning rate',
            'legend': False
        }

        self.assertEqual(target_train, value_train)
        np.testing.assert_array_almost_equal(target_val, value_val)
        self.assertDictEqual(target_kwargs, value_kwargs)
    """
    def test_get_best_epoch_info(self):
        self.history.summary_filename = self.summary_filename_2
        value = self.history.get_best_epoch_info()
        target = np.array(
            [(0.1, 0.11, 0.12, 0.13, 0.14, 0.15)],
            dtype=[('Epoch', '<f8'), ('LR', '<f8'), ('train_loss', '<f8'),
                   ('val_loss', '<f8'), ('train_acc', '<f8'), ('val_acc', '<f8')])
        np.testing.assert_array_equal(target, value)

    def test_get_best_epoch_fileno(self):
        self.history.summary_filename = self.summary_filename_2
        value = self.history.get_best_epoch_fileno()
        target = (1, 1)
        self.assertEqual(target, value)


def assert_equal_struc_array(a, b):
    """  np.testing.assert_array_equal does not work for arrays containing nans...
     so test individual instead. """
    np.testing.assert_array_equal(a.dtype, b.dtype)
    for name in a.dtype.names:
        np.testing.assert_almost_equal(a[name], b[name])



import tempfile
from unittest import TestCase
from unittest.mock import MagicMock
import os
import numpy as np
import tensorflow as tf

from orcanet.core import Organizer
from orcanet.h5_generator import get_h5_generator, _pad_to_size
from orcanet.tests.test_backend import save_dummy_h5py, assert_dict_arrays_equal, assert_equal_struc_array


class TestBatchGenerator(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tdir = tempfile.TemporaryDirectory()
        cls.temp_dir = cls.tdir.name

        # make some dummy data
        cls.n_bins = {'input_A': (2, 3), 'input_B': (2, 3)}
        cls.train_sizes = [5, 6]
        cls.train_A_file_1 = {
            "path": os.path.join(cls.temp_dir, "input_A_train_1.h5"),
            "shape": cls.n_bins["input_A"],
            "size": cls.train_sizes[0],
        }

        cls.train_B_file_1 = {
            "path":  os.path.join(cls.temp_dir, "input_B_train_1.h5"),
            "shape": cls.n_bins["input_B"],
            "size": cls.train_sizes[0],
        }

        cls.filepaths_file_1 = {
            "input_A": cls.train_A_file_1["path"],
            "input_B": cls.train_B_file_1["path"],
        }

        cls.train_A_file_1_ctnt = save_dummy_h5py(**cls.train_A_file_1)
        cls.train_B_file_1_ctnt = save_dummy_h5py(**cls.train_B_file_1)

    def setUp(self):
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.output_folder = self.data_folder + "/dummy_model"

        list_file = self.data_folder + "/in_out_test_list.toml"
        config_file = None

        self.orga = Organizer(self.output_folder, list_file, config_file)
        self.batchsize = 2
        self.orga.cfg.batchsize = self.batchsize
        self.orga.cfg.label_modifier = label_modifier

    @classmethod
    def tearDownClass(cls):
        cls.tdir.cleanup()

    def test_batch(self):
        filepaths = self.filepaths_file_1
        gene = iter(get_h5_generator(self.orga, filepaths))

        target_xs_batch_1 = {
            "input_A": self.train_A_file_1_ctnt[0][:2],
            "input_B": self.train_B_file_1_ctnt[0][:2],
        }

        target_ys_batch_1 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][:2]})

        target_xs_batch_2 = {
            "input_A": self.train_A_file_1_ctnt[0][2:4],
            "input_B": self.train_B_file_1_ctnt[0][2:4],
        }

        target_ys_batch_2 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][2:4]})

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)

        with self.assertRaises(StopIteration):
            next(gene)

    def test_batch_zero_center(self):
        filepaths = self.filepaths_file_1

        xs_mean = {name: np.ones(shape) * 0.5 for name, shape in self.n_bins.items()}

        self.orga.get_xs_mean = MagicMock(return_value=xs_mean)
        gene = iter(get_h5_generator(self.orga, filepaths, zero_center=True))

        target_xs_batch_1 = {
            "input_A": np.subtract(self.train_A_file_1_ctnt[0][:2], xs_mean["input_A"]),
            "input_B": np.subtract(self.train_B_file_1_ctnt[0][:2], xs_mean["input_B"]),
        }

        target_ys_batch_1 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][:2]})

        target_xs_batch_2 = {
            "input_A": np.subtract(self.train_A_file_1_ctnt[0][2:4], xs_mean["input_A"]),
            "input_B": np.subtract(self.train_B_file_1_ctnt[0][2:4], xs_mean["input_B"]),
        }

        target_ys_batch_2 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][2:4]})

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)

    def test_batch_sample_modifier(self):
        filepaths = self.filepaths_file_1

        def sample_modifier(info_blob):
            xs_in = info_blob["x_values"]
            mod = {name: val*2 for name, val in xs_in.items()}
            return mod

        self.orga.cfg.sample_modifier = sample_modifier
        gene = iter(get_h5_generator(self.orga, filepaths))

        target_xs_batch_1 = {
            "input_A": self.train_A_file_1_ctnt[0][:2]*2,
            "input_B": self.train_B_file_1_ctnt[0][:2]*2,
        }

        target_ys_batch_1 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][:2]})

        target_xs_batch_2 = {
            "input_A": self.train_A_file_1_ctnt[0][2:4]*2,
            "input_B": self.train_B_file_1_ctnt[0][2:4]*2,
        }

        target_ys_batch_2 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][2:4]})

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)

    def test_batch_mc_infos(self):
        filepaths = self.filepaths_file_1

        gene = iter(get_h5_generator(self.orga, filepaths, keras_mode=False))

        target_xs_batch_1 = {
            "input_A": self.train_A_file_1_ctnt[0][:2],
            "input_B": self.train_B_file_1_ctnt[0][:2],
        }

        target_ys_batch_1 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][:2]})
        target_mc_info_batch_1 = self.train_A_file_1_ctnt[1][:2]

        target_xs_batch_2 = {
            "input_A": self.train_A_file_1_ctnt[0][2:4],
            "input_B": self.train_B_file_1_ctnt[0][2:4],
        }

        target_ys_batch_2 = label_modifier({"y_values": self.train_A_file_1_ctnt[1][2:4]})
        target_mc_info_batch_2 = self.train_A_file_1_ctnt[1][2:4]

        info_blob = next(gene)
        assert_dict_arrays_equal(info_blob["xs"], target_xs_batch_1)
        assert_dict_arrays_equal(info_blob["ys"], target_ys_batch_1)
        assert_equal_struc_array(info_blob["y_values"], target_mc_info_batch_1)

        info_blob = next(gene)
        assert_dict_arrays_equal(info_blob["xs"], target_xs_batch_2)
        assert_dict_arrays_equal(info_blob["ys"], target_ys_batch_2)
        assert_equal_struc_array(info_blob["y_values"], target_mc_info_batch_2)

    def test_y_field_names(self):
        y_field_names = ("mc_A", )
        filepaths = self.filepaths_file_1
        self.orga.cfg.y_field_names = y_field_names
        gene = get_h5_generator(self.orga, filepaths, keras_mode=False)
        self.assertTupleEqual(gene[0]["y_values"].dtype.names, y_field_names)


class TestFuncs(tf.test.TestCase):
    def test_pad_to_size_with_tensors(self):
        x = tf.ones((5, 10))
        result = _pad_to_size(x, 7)
        self.assertShapeEqual(np.ones((7, 10)), result)

    def test_pad_to_size_with_tensors_shape_1(self):
        x = tf.ones((1, 10))
        result = _pad_to_size(x, 4)
        self.assertShapeEqual(np.ones((4, 10)), result)

    def test_pad_to_size_with_np_arrays(self):
        x = np.ones((5, 10))
        result = _pad_to_size(x, 7)
        self.assertTupleEqual((7, 10), result.shape)

    def test_pad_to_size_with_np_arrays_shape_1(self):
        x = np.ones((1, 10))
        result = _pad_to_size(x, 5)
        self.assertTupleEqual((5, 10), result.shape)


def label_modifier(info_blob):
    y_values = info_blob["y_values"]
    ys = dict()
    for name in y_values.dtype.names:
        ys[name] = y_values[name]
    return ys



"""
Run OrcaNet functionalities from command line.

"""
import argparse
from orcanet import version
# imports involving tf moved inside functions for speed up


def train(directory, list_file=None, config_file=None, model_file=None, to_epoch=None):
    from orcanet.core import Organizer
    from orcanet.model_builder import ModelBuilder
    from orcanet.misc import find_file

    orga = Organizer(directory, list_file, config_file, tf_log_level=1)

    if orga.io.get_latest_epoch() is None:
        # Start of training
        print("Building new model")
        if model_file is None:
            model_file = find_file(directory, "model.toml")
        model = ModelBuilder(model_file).build(orga, verbose=False)
    else:
        model = None

    return orga.train_and_validate(model=model, to_epoch=to_epoch)


def _add_parser_train(subparsers):
    parser = subparsers.add_parser(
        "train",
        description="Train and validate a model.",
    )
    _add_common_args(parser)
    parser.add_argument(
        "--model_file",
        type=str,
        help="Path to toml model file. Will be used to build a model at "
             "the start of the training. Not needed to resume training. "
             "Default: Look for a file called 'model.toml' in the "
             "given OrcaNet directory.",
        default=None,
    )
    parser.add_argument(
        "--to_epoch",
        type=int,
        help="Train up to and including this epoch. Default: Train forever.",
        default=None,
    )
    parser.set_defaults(func=train)


def predict(directory, list_file=None, config_file=None, epoch=None, fileno=None):
    from orcanet.core import Organizer

    orga = Organizer(directory, list_file, config_file, tf_log_level=1)
    return orga.predict(epoch=epoch, fileno=fileno)[0]


def _add_paser_predict(subparsers):
    parser = subparsers.add_parser(
        "predict",
        description="Load a trained model and save its prediction on "
                    "the predictions files to h5.",
    )
    _add_common_args(parser)
    parser.add_argument(
        "--epoch",
        type=int,
        help="Epoch of model to load. Default: best",
        default=None
    )
    parser.add_argument(
        "--fileno",
        type=int,
        help="Fileno of model to load. Default: best",
        default=None,
    )
    parser.set_defaults(func=predict)


def inference(directory, list_file=None, config_file=None, epoch=None, fileno=None):
    from orcanet.core import Organizer

    orga = Organizer(directory, list_file, config_file, tf_log_level=1)
    return orga.inference(epoch=epoch, fileno=fileno)


def _add_parser_inference(subparsers):
    parser = subparsers.add_parser(
        "inference",
        description="Load a trained model and save its prediction on the "
                    "inference files to h5.",
    )
    _add_common_args(parser)
    parser.add_argument(
        "--epoch",
        type=int,
        help="Epoch of model to load. Default: best",
        default=None,
    )
    parser.add_argument(
        "--fileno",
        type=int,
        help="Fileno of model to load. Default: best",
        default=None,
    )
    parser.set_defaults(func=inference)


def _add_common_args(prsr):
    prsr.add_argument(
        "directory",
        help="Path to OrcaNet directory.",
    )
    prsr.add_argument(
        "--list_file",
        type=str,
        help="Path to toml list file. Default: Look for a file called "
             "'list.toml' in the given OrcaNet directory.",
        default=None,
    )
    prsr.add_argument(
        "--config_file",
        type=str,
        help="Path to toml config file. Default: Look for a file called "
             "'config.toml' in the given OrcaNet directory.",
        default=None,
    )


def _add_parser_summarize(subparsers):
    import orcanet.utilities.summarize_training as summarize_training

    parent_parser = summarize_training.get_parser()
    parser = subparsers.add_parser(
        "summarize",
        description=parent_parser.description,
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parent_parser],
        add_help=False,
    )
    parser.set_defaults(func=summarize_training.summarize)


def main():
    parser = argparse.ArgumentParser(
        prog="orcanet",
        description=__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument('--version', action='version', version=version)

    subparsers = parser.add_subparsers()
    _add_parser_train(subparsers)
    _add_paser_predict(subparsers)
    _add_parser_inference(subparsers)
    _add_parser_summarize(subparsers)

    kwargs = vars(parser.parse_args())
    func = kwargs.pop("func")
    func(**kwargs)



import inspect
import warnings
import tensorflow.keras as ks
import tensorflow.keras.layers as layers

import orcanet.builder_util.layer_blocks as layer_blocks


class BlockBuilder:
    """
    Builds single-input block-wise sequential neural network.

    Parameters
    ----------
    defaults : dict or None
        Default values for all blocks in the model.
    verbose : bool
        Print info about the building process?
    batch_size : int, optional
        Define a fixed batchsize for the input.

    """
    def __init__(self, defaults=None, verbose=False, input_opts=None, **kwargs):
        """
        Set dict with default values for the layers of the model.
        Can also define custom block names as kwargs (key = toml name,
        value = block).
        """
        # dict with toml keyword vs block for all custom blocks
        self.all_blocks = {
            **layer_blocks.blocks,
            # legacy names:
            "conv_block": layer_blocks.ConvBlock,
            "dense_block": layer_blocks.DenseBlock,
            "resnet_block": layer_blocks.ResnetBlock,
            "resnet_bneck_block": layer_blocks.ResnetBnetBlock,
            "categorical": _attach_output_cat,
            "gpool": _attach_output_gpool_categ,
            "gpool_categ": _attach_output_gpool_categ,
            "gpool_reg": layer_blocks.OutputReg,
            "regression_error": layer_blocks.OutputRegErr,
        }

        if kwargs:
            self.all_blocks = {**self.all_blocks, **kwargs}

        self._check_arguments(defaults)
        self.defaults = defaults
        self.verbose = verbose
        if input_opts is None:
            self.input_opts = {}
        else:
            self.input_opts = input_opts

    def build(self, input_shape, configs):
        """
        Build the whole model, using the default values when arguments
        are missing in the layer_configs.

        Parameters
        ----------
        input_shape : dict
            Name and shape of the input layer.
        configs : list
            List of configurations for the blocks in the model.
            Each element in the list is a dict and will result in one block
            connected to the previous one. The dict has to contain the type
            of the block, as well as any arguments required by that
            specific block type.

        Returns
        -------
        model : keras model

        """
        input_layer = get_input_block(input_shape, **self.input_opts)

        x = input_layer
        for layer_config in configs:
            x = self.attach_block(x, layer_config)

        return ks.models.Model(inputs=input_layer, outputs=x)

    def attach_block(self, layer, layer_config):
        """
        Attach a block to the given layer based on the layer config.

        Will use the default values given during initialization if they are not
        present in the layer config.

        Parameters
        ----------
        layer : keras layer
            Layer to attach the block to.
        layer_config : dict
            Configuration of the block to attach. The dict has to contain
            the type of the block, as well as any arguments required by that
            specific block.

        Returns
        -------
        keras layer

        """
        filled = self._with_defaults(layer_config, self.defaults)
        if self.verbose:
            print(f"Attaching layer {filled} to tensor {layer}")
        block = self._get_blocks(filled.pop("type"))
        return block(**filled)(layer)

    def _with_defaults(self, config, defaults):
        """ Make a copy of a layer config and complete it with default values
        for its block, if they are missing in the layer config.
        """
        conf = dict(config)

        if config is not None and "type" in config:
            block_name = config["type"]
        elif defaults is not None and "type" in defaults:
            block_name = defaults["type"]
            conf["type"] = defaults["type"]
        else:
            raise KeyError("No layer block type specified")

        block = self._get_blocks(block_name)
        args = list(inspect.signature(block.__init__).parameters.keys())

        if defaults is not None:
            for key, val in defaults.items():
                if key in args and key not in conf:
                    conf[key] = val

        return conf

    def _get_blocks(self, name=None):
        """ Get the block class/function depending on the name. """
        if name is None:
            return self.all_blocks
        elif name.startswith("keras:"):
            return getattr(ks.layers, name.split("keras:")[1])
        elif name in self.all_blocks:
            return self.all_blocks[name]
        else:
            raise NameError(
                f"Unknown block type: {name}, must either start with "
                f"'keras:', or be one of {list(self.all_blocks.keys())}")

    def _check_arguments(self, defaults):
        """ Check if given defaults appear in at least one block. """
        if defaults is None:
            return
        # possible arguments for all blocks
        psb_args = ["type", ]
        for block in self._get_blocks().values():
            args = list(inspect.signature(block.__init__).parameters.keys())
            for arg in args:
                if arg not in psb_args and arg != "kwargs":
                    psb_args.append(arg)

        for t_def in defaults.keys():
            if t_def not in psb_args:
                warnings.warn(
                    f"Unknown default argument: {t_def} (has to appear in a block)")


def get_input_block(input_shapes, batchsize=None, names=None):
    """
    Build input layers according to a dict mapping the layer names to shapes.
    If none appears in shape, input is ragged.

    Parameters
    ----------
    input_shapes : dict
        Keys: Input layer names.
        Values: Their shapes.
    batchsize : int, optional
        Specify fixed batchsize.
    names : tuple, optional
        Make sure the inputs are these names and return them in this order.

    Returns
    -------
    inputs : tf.Tensor or tuple
        A list of named keras input layers, or the input Tensor if there
        is only one input.

    """
    if names is None:
        input_names = list(input_shapes.keys())
    else:
        if not set(names) == set(input_shapes.keys()):
            raise ValueError(f"Invalid input names: Expected {names} "
                             f"got {list(input_shapes.keys())}")
        input_names = names

    inputs = []
    for input_name in input_names:
        inputs.append(layers.Input(
            shape=input_shapes[input_name],
            name=input_name,
            dtype=ks.backend.floatx(),
            batch_size=batchsize,
            ragged=None in input_shapes[input_name],
        ))

    if len(inputs) == 1:
        return inputs[0]
    else:
        return tuple(inputs)


class _attach_output_cat:
    # legacy
    def __init__(self, categories, output_name,
                 flatten=True):
        self.categories = categories
        self.output_name = output_name
        self.flatten = flatten

    def __call__(self, layer):
        if self.flatten:
            transition = "keras:Flatten"
        else:
            transition = None
        out = layer_blocks.OutputCateg(
            categories=self.categories,
            output_name=self.output_name,
            transition=transition,
            unit_list=(128, 32),
        )(layer)
        return out


class _attach_output_gpool_categ:
    # legacy
    def __init__(self, categories, output_name, dropout=None):
        self.categories = categories
        self.output_name = output_name
        self.dropout = dropout

    def __call__(self, layer):
        x = layers.GlobalAveragePooling2D()(layer)
        if self.dropout is not None:
            x = layers.Dropout(self.dropout)(x)
        out = layer_blocks.OutputCateg(
            categories=self.categories,
            output_name=self.output_name,
            transition=None,
        )(x)
        return out



from unittest import TestCase
import tensorflow as tf

import orcanet.builder_util.layer_blocks as layer_blocks


class TestInceptionBlockV2(TestCase):
    def setUp(self):
        self.inp_2d = tf.keras.layers.Input(shape=(10, 10, 1))
        self.inp_3d = tf.keras.layers.Input(shape=(10, 10, 10, 1))

    def test_shape_2d(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
        )(self.inp_2d)
        self.assertSequenceEqual(
            tf.keras.Model(self.inp_2d, x).output_shape, (None, 10, 10, 21))

    def test_shape_2d_strides(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=2,
        )(self.inp_2d)
        self.assertSequenceEqual(
            tf.keras.Model(self.inp_2d, x).output_shape, (None, 5, 5, 15))

    def test_shape_3d(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=3,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=1,
        )(self.inp_3d)
        self.assertSequenceEqual(
            tf.keras.Model(self.inp_3d, x).output_shape, (None, 10, 10, 10, 21))

    def test_shape_3d_strides(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=3,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=2,
        )(self.inp_3d)
        self.assertSequenceEqual(
            tf.keras.Model(self.inp_3d, x).output_shape, (None, 5, 5, 5, 15))

    def test_2d_params(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
        )(self.inp_2d)
        self.assertEqual(
            tf.keras.Model(self.inp_2d, x).count_params(), 450)

    def test_2d_params_stride(self):
        x = layer_blocks.InceptionBlockV2(
            conv_dim=2,
            filters_pool=3,
            filters_1x1=4,
            filters_3x3=(5, 6),
            filters_3x3dbl=(7, 8),
            strides=2,
        )(self.inp_2d)
        self.assertEqual(
            tf.keras.Model(self.inp_2d, x).count_params(), 436)


class TestConvBlock(TestCase):
    def setUp(self):
        self.inp_2d = tf.keras.layers.Input(shape=(9, 10, 1))
        self.inp_3d = tf.keras.layers.Input(shape=(8, 9, 10, 1))

    def test_2d(self):
        x = layer_blocks.ConvBlock(
            conv_dim=2,
            filters=11,
            dropout=0.2,
            pool_size=2,
        )(self.inp_2d)
        model = tf.keras.Model(self.inp_2d, x)

        self.assertEqual(model.count_params(), 110)
        self.assertSequenceEqual(model.output_shape, (None, 4, 5, 11))
        target_layers = (
            tf.keras.layers.InputLayer,
            tf.keras.layers.Conv2D,
            tf.keras.layers.Activation,
            tf.keras.layers.MaxPool2D,
            tf.keras.layers.Dropout,
        )
        for layer, target_layer in zip(model.layers, target_layers):
            self.assertIsInstance(layer, target_layer)

    def test_2d_time_distributed(self):
        x = layer_blocks.ConvBlock(
            conv_dim=2,
            filters=11,
            dropout=0.2,
            pool_size=2,
            time_distributed=True,
            padding=[1, 0],
        )(self.inp_3d)
        model = tf.keras.Model(self.inp_3d, x)

        self.assertEqual(model.count_params(), 110)
        self.assertSequenceEqual(model.output_shape, (None, 8, 4, 4, 11))
        target_layers = [tf.keras.layers.InputLayer] + [tf.keras.layers.TimeDistributed] * 5

        for layer, target_layer in zip(model.layers, target_layers):
            self.assertIsInstance(layer, target_layer)


class TestOutputRegNormal(TestCase):
    # also using this class as template for other tests
    @classmethod
    def setUpClass(cls):
        inp = tf.keras.layers.Input(shape=(3, 4))
        output = layer_blocks.OutputRegNormal(
            output_neurons=3,
            output_name="test",
            unit_list=[5, 5],
            mu_activation="relu",
            transition="keras:Flatten",
        )(inp)
        cls.model = tf.keras.Model(inp, output)

    def setUp(self):
        self.targets = {
            "n_layers": 11,
            "output_shape": (None, 2, 3),
            "n_params": 131,
            "output_names": ["test"],
        }

    def test_unit_list_is_int(self):
        inp = tf.keras.layers.Input(shape=(3, 4))
        output = layer_blocks.OutputRegNormal(
            output_neurons=3,
            output_name="test",
            unit_list=5,
            mu_activation="relu",
            transition="keras:Flatten",
        )(inp)
        model = tf.keras.Model(inp, output)

    def test_n_layers(self):
        self.assertEqual(len(self.model.layers), self.targets["n_layers"])

    def test_output_shape(self):
        self.assertTupleEqual(self.model.output_shape, self.targets["output_shape"])

    def test_n_params(self):
        self.assertEqual(self.model.count_params(), self.targets["n_params"])

    def test_output_names(self):
        if "output_names" in self.targets:
            self.assertListEqual(self.model.output_names, self.targets["output_names"])


class TestOutputRegNormalSplit(TestOutputRegNormal):
    @classmethod
    def setUpClass(cls):
        inp = tf.keras.layers.Input(shape=(3, 4))
        output = layer_blocks.OutputRegNormalSplit(
            output_neurons=3,
            output_name="test",
            unit_list=[5, 5],
            sigma_unit_list=[3, ],
            mu_activation="relu",
            transition="keras:Flatten",
        )(inp)
        cls.model = tf.keras.Model(inp, output)

    def setUp(self):
        self.targets = {
            "n_layers": 15,
            "output_shape": [(None, 3), (None, 2, 3)],
            "n_params": 164,
            "output_names": ["test", "test_err"],
        }

    def test_output_shape(self):
        for i in range(len(self.targets["output_shape"])):
            self.assertTupleEqual(self.model.output_shape[i], self.targets["output_shape"][i])


class TestResnetBnetBlock(TestOutputRegNormal):
    @classmethod
    def setUpClass(cls):
        inp = tf.keras.layers.Input(shape=(5, 4, 3, 2))
        output = layer_blocks.ResnetBnetBlock(
            conv_dim=2,
            filters=[2, 3, 4],
            strides=2,
            batchnorm=True,
        )(inp)
        cls.model = tf.keras.Model(inp, output)

    def setUp(self):
        self.targets = {
            "n_layers": 13,
            "output_shape": (None, 5, 2, 2, 4),
            "n_params": 130,
        }



import numpy as np
import h5py
from tensorflow.keras.layers import Input, Dense
from tensorflow.keras.models import Model
import os

from orcanet.core import Organizer


def make_dummy_data(path):
    """
    Save a train and a val h5 file with random numbers as samples,
    and the sum as labels.

    """
    def get_dummy_data(samples):
        xs = np.random.rand(samples, 3)

        dtypes = [('sum', '<f8'), ]
        labels = np.sum(xs, axis=-1)
        ys = labels.ravel().view(dtype=dtypes)

        return xs, ys

    def save_dummy_data(path, samples):
        xs, ys = get_dummy_data(samples)

        with h5py.File(path, 'w') as h5f:
            h5f.create_dataset('x', data=xs, dtype='<f8')
            h5f.create_dataset('y', data=ys, dtype=ys.dtype)
    save_dummy_data(path + "example_train.h5", 40000)
    save_dummy_data(path + "example_val.h5", 5000)


def make_dummy_model():
    """
    Build and compile a small dummy model.
    """

    input_shape = (3,)

    inp = Input(input_shape, name="random_numbers")
    x = Dense(10)(inp)
    outp = Dense(1, name="sum")(x)

    model = Model(inp, outp)
    model.compile("sgd", loss="mae")

    return model


def use_orcanet():
    temp_folder = "output/"
    os.mkdir(temp_folder)

    make_dummy_data(temp_folder)
    list_file = "example_list.toml"

    organizer = Organizer(temp_folder + "sum_model", list_file)
    organizer.cfg.train_logger_display = 10

    model = make_dummy_model()
    organizer.train_and_validate(model, epochs=3)

    organizer.predict()


use_orcanet()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Core scripts for the OrcaNet package.
"""

import os
import toml
import warnings
import time
from datetime import timedelta
import tensorflow as tf

import orcanet.backend as backend
from orcanet.utilities.visualization import update_summary_plot
from orcanet.in_out import IOHandler
from orcanet.history import HistoryHandler
from orcanet.utilities.nn_utilities import load_zero_center_data
import orcanet.lib as lib
import orcanet.logging as logging
import orcanet.misc
import medgeconv


class Organizer:
    """
    Core class for working with networks in OrcaNet.

    Attributes
    ----------
    cfg : orcanet.core.Configuration
        Contains all configurable options.
    io : orcanet.in_out.IOHandler
        Utility functions for accessing the info in cfg.
    history : orcanet.in_out.HistoryHandler
        For reading and plotting data from the log files created
        during training.

    """
    def __init__(self, output_folder,
                 list_file=None,
                 config_file=None,
                 tf_log_level=None,
                 discover_tomls=True):
        """
        Set the attributes of the Configuration object.

        Instead of using a config_file, the attributes of orga.cfg can
        also be changed directly, e.g. by calling orga.cfg.batchsize.

        Parameters
        ----------
        output_folder : str
            Name of the folder of this model in which everything will be saved,
            e.g., the summary.txt log file is located in here.
            Will be used to load saved files or to save new ones.
        list_file : str, optional
            Path to a toml list file with pathes to all the h5 files that should
            be used for training and validation.
            Will be used to extract samples and labels.
            Default: Look for a file called 'list.toml' in the given output_folder.
        config_file : str, optional
            Path to a toml config file with settings that are used instead of
            the default ones.
            Default: Look for a file called 'config.toml' in the given output_folder.
        tf_log_level : int/str
            Sets the TensorFlow CPP_MIN_LOG_LEVEL environment variable.
            0 = all messages are logged (default behavior).
            1 = INFO messages are not printed.
            2 = INFO and WARNING messages are not printed.
            3 = INFO, WARNING, and ERROR messages are not printed.
        discover_tomls : bool
            If False, do not try to look for toml files in the given
            output_folder if list_file or config_file is None [Default: True].

        """
        if tf_log_level is not None:
            os.environ['TF_CPP_MIN_LOG_LEVEL'] = str(tf_log_level)

        if discover_tomls and list_file is None:
            list_file = orcanet.misc.find_file(output_folder, "list.toml")
        if discover_tomls and config_file is None:
            config_file = orcanet.misc.find_file(output_folder, "config.toml")

        self.cfg = Configuration(output_folder, list_file, config_file)
        self.io = IOHandler(self.cfg)
        self.history = HistoryHandler(output_folder)

        self.xs_mean = None
        self._auto_label_modifier = None
        self._stored_model = None
        self._strategy = None

    def train_and_validate(self, model=None, epochs=None, to_epoch=None):
        """
        Train a model and validate according to schedule.

        The various settings of this process can be controlled with the
        attributes of orca.cfg.
        The model will be trained on the given data, saved and validated.
        Logfiles of the training are saved in the output folder.
        Plots showing the training and validation history, as well as
        the weights and activations of the network are generated in
        the plots subfolder after every validation.
        The training can be resumed by executing this function again.

        Parameters
        ----------
        model : ks.models.Model or str, optional
            Compiled keras model to use for training. Required for the first
            epoch (the start of training).
            Can also be the path to a saved keras model, which will be laoded.
            If model is None, the most recent saved model will be
            loaded automatically to continue the training.
        epochs : int, optional
            How many epochs should be trained by running this function.
            None for infinite. This includes the current epoch in case it
            is not finished yet, i.e. 1 means complete the epoch if there
            are files left, otherwise do the next epoch.
        to_epoch : int, optional
            Train up to and including this epoch. Can not be used together with
            epochs.

        Returns
        -------
        model : ks.models.Model
            The trained keras model.

        """
        latest_epoch = self.io.get_latest_epoch()

        model = self._get_model(model, logging=False)
        self._stored_model = model

        # check if the validation is missing for the latest fileno
        if latest_epoch is not None:
            state = self.history.get_state()[-1]
            if state["is_validated"] is False and self.val_is_due(latest_epoch):
                self.validate()

        next_epoch = self.io.get_next_epoch(latest_epoch)
        n_train_files = self.io.get_no_of_files("train")

        if to_epoch is None:
            epochs_left = epochs
        else:
            if epochs is not None:
                raise ValueError("Can not give both 'epochs' and 'to_epoch'")
            if latest_epoch is None:
                epochs_left = to_epoch
            else:
                epochs_left = max(
                    0, to_epoch - self.io.get_next_epoch(latest_epoch)[0] + 1)

        trained_epochs = 0
        while epochs_left is None or trained_epochs < epochs_left:
            # Train on remaining files
            for file_no in range(next_epoch[1], n_train_files + 1):
                curr_epoch = (next_epoch[0], file_no)
                self.train(model)
                if self.val_is_due(curr_epoch):
                    self.validate()

            next_epoch = (next_epoch[0] + 1, 1)
            trained_epochs += 1

        self._stored_model = None
        return model

    def train(self, model=None):
        """
        Trains a model on the next file.

        The progress of the training is also logged and plotted.

        Parameters
        ----------
        model : ks.models.Model or str, optional
            Compiled keras model to use for training. Required for the first
            epoch (the start of training).
            Can also be the path to a saved keras model, which will be laoded.
            If model is None, the most recent saved model will be
            loaded automatically to continue the training.

        Returns
        -------
        history : dict
            The history of the training on this file. A record of training
            loss values and metrics values.

        """
        # Create folder structure
        self.io.get_subfolder(create=True)
        latest_epoch = self.io.get_latest_epoch()

        model = self._get_model(model, logging=True)

        self._set_up(model, logging=True)

        # epoch about to be trained
        next_epoch = self.io.get_next_epoch(latest_epoch)
        next_epoch_float = self.io.get_epoch_float(*next_epoch)

        if latest_epoch is None:
            self.io.check_connections(model)
            logging.log_start_training(self)

        model_path = self.io.get_model_path(*next_epoch)
        model_path_local = self.io.get_model_path(*next_epoch, local=True)
        if os.path.isfile(model_path):
            raise FileExistsError(
                "Can not train model in epoch {} file {}, this model has "
                "already been saved!".format(*next_epoch))

        smry_logger = logging.SummaryLogger(self, model)

        if self.cfg.learning_rate is not None:
            tf.keras.backend.set_value(
                model.optimizer.lr, self.io.get_learning_rate(next_epoch)
            )

        files_dict = self.io.get_file("train", next_epoch[1])

        line = "Training in epoch {} on file {}/{}".format(
            next_epoch[0], next_epoch[1], self.io.get_no_of_files("train"))
        self.io.print_log(line)
        self.io.print_log("-" * len(line))
        self.io.print_log("Learning rate is at {}".format(
            tf.keras.backend.get_value(model.optimizer.lr)))
        self.io.print_log('Inputs and files:')
        for input_name, input_file in files_dict.items():
            self.io.print_log("   {}: \t{}".format(input_name,
                                                   os.path.basename(
                                                       input_file)))

        start_time = time.time()
        history = backend.train_model(self, model, next_epoch, batch_logger=True)
        elapsed_s = int(time.time() - start_time)

        model.save(model_path)
        smry_logger.write_line(
            next_epoch_float,
            tf.keras.backend.get_value(model.optimizer.lr),
            history_train=history,
        )

        self.io.print_log('Training results:')
        for metric_name, loss in history.items():
            self.io.print_log(f"   {metric_name}: \t{loss}")
        self.io.print_log(f"Elapsed time: {timedelta(seconds=elapsed_s)}")
        self.io.print_log(f"Saved model to: {model_path_local}\n")

        update_summary_plot(self)
        if self.cfg.cleanup_models:
            self.cleanup_models()

        return history

    def validate(self):
        """
        Validate the most recent saved model on all validation files.

        Will also log the progress, as well as update the summary plot and
        plot weights and activations of the model.

        Returns
        -------
        history : dict
            The history of the validation on all files. A record of validation
            loss values and metrics values.

        """
        latest_epoch = self.io.get_latest_epoch()
        if latest_epoch is None:
            raise ValueError("Can not validate: No saved model found")
        if self.history.get_state()[-1]["is_validated"] is True:
            raise ValueError("Can not validate in epoch {} file {}: "
                             "Has already been validated".format(*latest_epoch))

        if self._stored_model is None:
            model = self.load_saved_model(*latest_epoch)
        else:
            model = self._stored_model

        self._set_up(model, logging=True)

        epoch_float = self.io.get_epoch_float(*latest_epoch)
        smry_logger = logging.SummaryLogger(self, model)

        logging.log_start_validation(self)

        start_time = time.time()
        history = backend.validate_model(self, model)
        elapsed_s = int(time.time() - start_time)

        self.io.print_log('Validation results:')
        for metric_name, loss in history.items():
            self.io.print_log(f"   {metric_name}: \t{loss}")
        self.io.print_log(f"Elapsed time: {timedelta(seconds=elapsed_s)}\n")
        smry_logger.write_line(epoch_float, "n/a", history_val=history)

        update_summary_plot(self)

        if self.cfg.cleanup_models:
            self.cleanup_models()

        return history

    def predict(self, epoch=None, fileno=None, samples=None):
        """
        Make a prediction if it does not exist yet, and return its filepath.

        Load the model with the lowest validation loss, let it predict on
        all samples of the validation set
        in the toml list, and save this prediction together with all the
        y_values as h5 file(s) in the predictions subfolder.

        Parameters
        ----------
        epoch : int, optional
            Epoch of a model to load. Default: lowest val loss.
        fileno : int, optional
            File number of a model to load. Default: lowest val loss.
        samples : int, optional
            Don't use the full validation files, but just the given number
            of samples.

        Returns
        -------
        pred_filename : List
            List to the paths of all the prediction files.

        """
        if fileno is None and epoch is None:
            epoch, fileno = self.history.get_best_epoch_fileno()
            print(f"Automatically set epoch to epoch {epoch} file {fileno}.")
        elif fileno is None or epoch is None:
            raise ValueError(
                "Either both or none of epoch and fileno must be None")

        if self._check_if_pred_already_done(epoch, fileno):
            print("Prediction has already been done.")
            pred_filepaths = self.io.get_pred_files_list(epoch, fileno)

        else:
            if self._stored_model is None:
                model = self.load_saved_model(epoch, fileno, logging=False)
            else:
                model = self._stored_model
            self._set_up(model)

            start_time = time.time()
            backend.make_model_prediction(
                self, model, epoch, fileno, samples=samples)
            elapsed_s = int(time.time() - start_time)
            print('Finished predicting on all validation files.')
            print("Elapsed time: {}\n".format(timedelta(seconds=elapsed_s)))

            pred_filepaths = self.io.get_pred_files_list(epoch, fileno)

        return pred_filepaths

    def inference(self, epoch=None, fileno=None, as_generator=False):
        """
        Make an inference and return the filepaths.

        Load the model with the lowest validation loss, let
        it predict on all samples of all inference files
        in the toml list, and save these predictions as h5 files in the
        predictions subfolder. y values will only be added if they are in
        the input file, so this can be used on un-labeled data as well.

        Parameters
        ----------
        epoch : int, optional
            Epoch of a model to load [default: lowest val loss].
        fileno : int, optional
            File number of a model to load [default: lowest val loss].
        as_generator : bool
            If true, return a generator, which yields the output filename
            after the inference of each file.
            If false (default), do all files back to back.

        Returns
        -------
        filenames : list
            List to the paths of all created output files.

        """
        gen = self._inference(epoch=epoch, fileno=fileno)
        if as_generator:
            return gen
        else:
            return [filename for filename in gen]

    def _inference(self, epoch=None, fileno=None):
        """ Get generator of doing inference, file by file. """
        if fileno is None and epoch is None:
            epoch, fileno = self.history.get_best_epoch_fileno()
            print("Automatically set epoch to epoch {} file {}.".format(epoch, fileno))
        elif fileno is None or epoch is None:
            raise ValueError(
                "Either both or none of epoch and fileno must be None")

        if self._stored_model is None:
            model = self.load_saved_model(epoch, fileno, logging=False)
        else:
            model = self._stored_model
        self._set_up(model)

        filenames = []
        for files_dict in self.io.yield_files("inference"):
            # output filename is based on name of file in first input
            first_filename = os.path.basename(list(files_dict.values())[0])
            output_filename = "model_epoch_{}_file_{}_on_{}".format(
                epoch, fileno, first_filename)

            output_path = os.path.join(self.io.get_subfolder("inference"),
                                       output_filename)
            filenames.append(output_path)
            if os.path.exists(output_path):
                print("File {} exists already, skipping...".format(output_filename))
                continue

            print(f'Working on file {first_filename}')
            start_time = time.time()
            backend.h5_inference(
                self, model, files_dict, output_path, use_def_label=False)
            elapsed_s = int(time.time() - start_time)
            print(f'Finished on file {first_filename} in {elapsed_s/60} min')
            yield output_path

    def cleanup_models(self):
        """
        Delete all models except for the the most recent one (to continue
        training), and the ones with the highest and lowest loss/metrics.

        """
        all_epochs = self.io.get_all_epochs()
        epochs_to_keep = {self.io.get_latest_epoch(), }
        try:
            for metric in self.history.get_metrics():
                epochs_to_keep.add(
                    self.history.get_best_epoch_fileno(
                        metric=f"val_{metric}", mini=True))
                epochs_to_keep.add(
                    self.history.get_best_epoch_fileno(
                        metric=f"val_{metric}", mini=False))
        except ValueError:
            # no best epoch exists
            pass

        for epoch in epochs_to_keep:
            if epoch not in all_epochs:
                warnings.warn(
                    f"ERROR: keeping_epoch {epoch} not in available epochs {all_epochs}, "
                    f"skipping clean-up of models!")
                return

        print("\nClean-up saved models:")
        for epoch in all_epochs:
            model_path = self.io.get_model_path(epoch[0], epoch[1])
            model_name = os.path.basename(model_path)
            if epoch in epochs_to_keep:
                print("Keeping model {}".format(model_name))
            else:
                print("Deleting model {}".format(model_name))
                os.remove(model_path)

    def _check_if_pred_already_done(self, epoch, fileno):
        """
        Checks if the prediction has already been done before.
        (-> predicted on all validation files)

        Returns
        -------
        pred_done : bool
            Boolean flag to specify if the prediction has
            already been fully done or not.

        """
        latest_pred_file_no = self.io.get_latest_prediction_file_no(epoch, fileno)
        total_no_of_val_files = self.io.get_no_of_files('val')

        if latest_pred_file_no is None:
            pred_done = False
        elif latest_pred_file_no == total_no_of_val_files:
            return True
        else:
            pred_done = False

        return pred_done

    def get_xs_mean(self, logging=False):
        """
        Set and return the zero center image for each list input.

        Requires the cfg.zero_center_folder to be set. If no existing
        image for the given input files is found in the folder, it will
        be calculated and saved by averaging over all samples in the
        train dataset.

        Parameters
        ----------
        logging : bool
            If true, the execution of this function will be logged into the
            full summary in the output folder if called for the first time.

        Returns
        -------
        dict
            Dict of numpy arrays that contains the mean_image of the x dataset
            (1 array per list input).
            Example format:
            { "input_A" : ndarray, "input_B" : ndarray }

        """
        if self.xs_mean is None:
            if self.cfg.zero_center_folder is None:
                raise ValueError("Can not calculate zero center: "
                                 "No zero center folder given")
            self.xs_mean = load_zero_center_data(self, logging=logging)
        return self.xs_mean

    def load_saved_model(self, epoch, fileno, logging=False):
        """
        Load a saved model.

        Parameters
        ----------
        epoch : int
            Epoch of the saved model. If both this and fileno are -1,
            load the most recent model.
        fileno : int
            Fileno of the saved model.
        logging : bool
            If True, will log this function call into the log.txt file.

        Returns
        -------
        model : keras model

        """
        path_of_model = self.io.get_model_path(epoch, fileno)
        path_loc = self.io.get_model_path(epoch, fileno, local=True)
        self.io.print_log("Loading saved model: " + path_loc, logging=logging)
        return self._load_model(path_of_model)

    def _get_model(self, model, logging=False):
        """ Load most recent saved model or use user model. """
        latest_epoch = self.io.get_latest_epoch()

        if latest_epoch is None:
            # new training, log info about model
            if model is None:
                raise ValueError("You need to provide a compiled keras model "
                                 "for the start of the training! (You gave None)")

            elif isinstance(model, str):
                # path to a saved model
                self.io.print_log("Loading model from " + model, logging=logging)
                model = self._load_model(model)

            if logging:
                self._save_as_json(model)
                model.summary(print_fn=self.io.print_log)

                try:
                    plots_folder = self.io.get_subfolder("plots", create=True)
                    tf.keras.utils.plot_model(
                        model, plots_folder + "/model_plot.png", show_shapes=True)
                except (ImportError, AttributeError) as e:
                    # TODO remove AttributeError once https://github.com/tensorflow/tensorflow/issues/38988 is fixed
                    warnings.warn("Can not plot model: " + str(e))

        else:
            # resuming training, load model if it is not given
            if model is None:
                model = self.load_saved_model(*latest_epoch, logging=logging)

            elif isinstance(model, str):
                # path to a saved model
                self.io.print_log("Loading model from " + model, logging=logging)
                model = self._load_model(model)

        return model

    def _load_model(self, filepath):
        """ Load from path, with custom objects and parallized. """
        with self.get_strategy().scope():
            model = tf.keras.models.load_model(
                filepath, custom_objects=self.cfg.get_custom_objects())
        return model

    def _save_as_json(self, model):
        """ Save the architecture of a model as json to fixed path. """
        json_filename = "model_arch.json"

        json_string = model.to_json(indent=1)
        model_folder = self.io.get_subfolder("saved_models", create=True)
        with open(os.path.join(model_folder, json_filename), "w") as f:
            f.write(json_string)

    def _set_up(self, model, logging=False):
        """ Necessary setup for training, validating and predicting. """
        if self.cfg.get_list_file() is None:
            raise ValueError("No files specified. Need to load a toml "
                             "list file with pathes to h5 files first.")

        if self.cfg.label_modifier is None:
            self._auto_label_modifier = lib.label_modifiers.ColumnLabels(model)

        if self.cfg.zero_center_folder is not None:
            self.get_xs_mean(logging)

    def val_is_due(self, epoch=None):
        """
        True if validation is due on given epoch according to schedule.
        Does not check if it has been done already.

        """
        if epoch is None:
            epoch = self.io.get_latest_epoch()
        n_train_files = self.io.get_no_of_files("train")
        val_sched = (epoch[1] == n_train_files) or \
                    (self.cfg.validate_interval is not None and
                     epoch[1] % self.cfg.validate_interval == 0)
        return val_sched

    def get_strategy(self):
        """ Get the strategy for distributed training. """
        if self._strategy is None:
            if self.cfg.multi_gpu and len(
                    tf.config.list_physical_devices('GPU')) > 1:
                self._strategy = tf.distribute.MirroredStrategy()
                print(f'Number of GPUs: {self._strategy.num_replicas_in_sync}')
            else:
                self._strategy = tf.distribute.get_strategy()
        return self._strategy


class Configuration(object):
    """
    Contains all the configurable options in the OrcaNet scripts.

    All of these public attributes (the ones without a
    leading underscore) can be changed either directly or with a
    .toml config file via the method update_config().

    Parameters
    ----------
    output_folder : str
        Name of the folder of this model in which everything will be saved,
        e.g., the summary.txt log file is located in here.
    list_file : str or None
        Path to a toml list file with pathes to all the h5 files that should
        be used for training and validation.
    config_file : str or None
        Path to a toml config file with attributes that are used instead of
        the default ones.
    kwargs
        Overwrites the values given in the config file.

    Attributes
    ----------
    batchsize : int
        Batchsize that will be used for the training, validation and inference of
        the network.
        During training and validation, the last batch in each file will be
        skipped if it has fewer samples than the batchsize.
    callback_train : keras callback or list or None
        Callback or list of callbacks to use during training.
    class_weight : dict or None
        Optional dictionary mapping class indices (integers) to a weight
        (float) value, used for weighting the loss function (during
        training only). This can be useful to tell the model to
        "pay more attention" to samples from an under-represented class.
    cleanup_models : bool
        If true, will only keep the best (in terms of val loss) and the most
        recent from all saved models in order to save disk space.
    custom_objects : dict, optional
        Optional dictionary mapping names (strings) to custom classes or
        functions to be considered by keras during deserialization of models.
    dataset_modifier : function or None
        For orga.predict: Function that determines which datasets get created
        in the resulting h5 file. Default: save as array, i.e. every output layer
        will get one dataset each for both the label and the prediction,
        and one dataset containing the y_values from the validation files.
    fixed_batchsize : bool
        The last batch in the file might be smaller then the batchsize.
        Usually, this is no problem, but set to True to skip this batch
        [default: False].
    key_x_values : str
        The name of the datagroup in the h5 input files which contains
        the samples for the network.
    key_y_values : str
        The name of the datagroup in the h5 input files which contains
        the info for the labels.
    label_modifier : function or None
        Operation to be performed on batches of y_values read from the input
        files before they are fed into the model as labels. If None is given,
        all y_values with the same name as the output layers will be passed
        to the model as a dict, with the keys being the dtype names.
    learning_rate : float, tuple, function, str (optional)
        The learning rate for the training.
        If None is given, don't change the learning rate at all.
        If it is a float: The learning rate will be constantly this value.
        If it is a tuple of two floats: The first float gives the learning rate
        in epoch 1 file 1, and the second float gives the decrease of the
        learning rate per file (e.g. 0.1 for 10% decrease per file).
        If it is a function: Takes as an input the epoch and the
        file number (in this order), and returns the learning rate.
        Both epoch and fileno start at 1, i.e. 1, 1 is the start of the
        training.
        If it is a str: Path to a csv file inside the main folder, containing
        3 columns with the epoch, fileno, and the value the lr will be set
        to when reaching this epoch/fileno.
    max_queue_size : int
        max_queue_size option of the keras training and evaluation generator
        methods. How many batches get preloaded from the generator.
    multi_gpu : bool
        Use all availble GPUs (distributed training if theres more then one).
    n_events : None or int
        For testing purposes. If not the whole .h5 file should be used for
        training, define the number of samples.
    sample_modifier : function or None
        Operation to be performed on batches of x_values read from the input
        files before they are fed into the model as samples.
    shuffle_train : bool
        If true, the order in which batches are read out from the files during
        training are randomized each time they are read out.
    train_logger_display : int
        How many batches should be averaged for one line in the training log files.
    train_logger_flush : int
        After how many lines the training log file should be flushed (updated on
        the disk). -1 for flush at the end of the file only.
    use_scratch_ssd : bool
        Declares if the input files should be copied to a local temp dir,
        i.e. the path defined in the 'TMPDIR' environment variable.
    validate_interval : int or None
        Validate the model after this many training files have been trained on
        in an epoch. There will always be a validation at the end of an epoch.
        None for only validate at the end of an epoch.
        Example: validate_interval=3 --> Validate after file 3, 6, 9, ...
    verbose_train : int
        verbose option of keras.model.fit_generator.
        0 = silent, 1 = progress bar, 2 = one line per epoch.
    verbose_val : int
        verbose option of evaluate_generator.
        0 = silent, 1 = progress bar.
    y_field_names : tuple or list or str, optional
        During train and val, read out only these fields from the y dataset.
        --> Speed up, especially if there are many fields.
    zero_center_folder : None or str
        Path to a folder in which zero centering images are stored.
        If this path is set, zero centering images for the given dataset will
        either be calculated and saved automatically at the start of the
        training, or loaded if they have been saved before.

    """
    # TODO add a clober script that properly deletes models + logfiles
    def __init__(self, output_folder, list_file=None, config_file=None, **kwargs):
        self.batchsize = 64
        self.callback_train = []
        self.class_weight = None
        self.cleanup_models = False
        self.custom_objects = {}
        self.dataset_modifier = None
        self.fixed_batchsize = False
        self.key_x_values = "x"
        self.key_y_values = "y"
        self.label_modifier = None
        self.learning_rate = None
        self.make_weight_plots = False  # Removed in v0.11.1
        self.max_queue_size = 10
        self.multi_gpu = True
        self.n_events = None
        self.sample_modifier = None
        self.shuffle_train = False
        self.train_logger_display = 100
        self.train_logger_flush = -1
        self.use_scratch_ssd = False
        self.validate_interval = None
        self.verbose_train = 1
        self.verbose_val = 0
        self.y_field_names = None
        self.zero_center_folder = None

        self._default_values = dict(self.__dict__)

        # Main folder:
        if output_folder[-1] == "/":
            self.output_folder = output_folder
        else:
            self.output_folder = output_folder+"/"

        # Private attributes:
        self._files_dict = {
            "train": None,
            "val": None,
            "inference": None,
        }
        self._list_file = None

        # Load the optionally given list and config files.
        if list_file is not None:
            self.import_list_file(list_file)
        if config_file is not None:
            self.update_config(config_file)

        # set given kwargs:
        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                raise AttributeError(
                    "Unknown attribute {}".format(key))

        # deprecation warning TODO remove in the future
        if self.make_weight_plots:
            warnings.warn("make_weight_plots was removed in version v0.11.1")

    def import_list_file(self, list_file):
        """
        Import the filepaths of the h5 files from a toml list file.

        Parameters
        ----------
        list_file : str
            Path to the toml list file.

        """
        if self._list_file is not None:
            raise ValueError("Can not load list file: Has already been loaded! "
                             "({})".format(self._list_file))

        file_content = toml.load(list_file)

        name_mapping = {
            "train_files": "train",
            "validation_files": "val",
            "inference_files": "inference",
        }

        for toml_name, files_dict_name in name_mapping.items():
            files = _extract_filepaths(file_content, toml_name)
            self._files_dict[files_dict_name] = files or None

        self._list_file = list_file

    def update_config(self, config_file):
        """
        Update the default cfg parameters with values from a toml config file.

        Parameters
        ----------
        config_file : str
            Path to a toml config file.

        """
        user_values = toml.load(config_file)["config"]
        for key, value in user_values.items():
            if hasattr(self, key):
                if key == "sample_modifier":
                    value = orcanet.misc.from_register(
                        toml_entry=value, register=lib.sample_modifiers.smods)
                elif key == "dataset_modifier":
                    value = orcanet.misc.from_register(
                        toml_entry=value, register=lib.dataset_modifiers.dmods)
                elif key == "label_modifier":
                    value = orcanet.misc.from_register(
                        toml_entry=value, register=lib.label_modifiers.lmods)
                setattr(self, key, value)
            else:
                raise AttributeError(f"Unknown attribute {key} in config file")

    def get_list_file(self):
        """
        Returns the path to the list file that was used to set the training
        and validation files. None if no list file has been used.

        """
        return self._list_file

    def get_files(self, which):
        """
        Get the training or validation file paths for each list input set.

        Parameters
        ----------
        which : str
            Either "train", "val" or "inference".

        Returns
        -------
        dict
            A dict containing the paths to the training or validation files on
            which the model will be trained on. Example for the format for
            two input sets with two files each:
                    {
                     "input_A" : ('path/to/set_A_file_1.h5', 'path/to/set_A_file_2.h5'),
                     "input_B" : ('path/to/set_B_file_1.h5', 'path/to/set_B_file_2.h5'),
                    }

        """
        if which not in self._files_dict.keys():
            raise NameError("Unknown fileset name ", which)
        if self._files_dict[which] is None:
            raise AttributeError("No {} files have been specified!".format(which))
        return self._files_dict[which]

    def get_custom_objects(self):
        """ Get user custom objects + orcanet internal ones. """
        orcanet_co = medgeconv.custom_objects
        orcanet_loss_functions = lib.losses.loss_functions
        return {**orcanet_co, **orcanet_loss_functions, **self.custom_objects}


def _get_h5_files(folder):
    h5files = []
    for f in os.listdir(folder):
        if f.endswith(".h5"):
            h5files.append(os.path.join(folder, f))
    h5files.sort()
    if not h5files:
        warnings.warn(f"No .h5 files in dir {folder}!")
    return h5files


def _extract_filepaths(file_content, which):
    """
    Get train, val or inf filepaths of all inputs from a toml readout.
    Makes sure that all input have the same number of files.

    """
    # alternative names to write in the toml file
    aliases = {
        "train_files": ("training_files", "train", "training"),
        "validation_files": ("val_files", "val", "validation"),
        "inference_files": ("inf_files", "inf", "inference"),
    }
    assert which in aliases.keys(), f"{which} not in {list(aliases.keys())}"

    def get_alias(ident):
        for k, v in aliases.items():
            if ident == k or ident in v:
                return k
        else:
            raise NameError(
                f"Unknown argument '{ident}' in toml file: "
                f"Must be either of {list(aliases.keys())}"
            )

    files = {}
    n_files = []
    for input_name, input_files in file_content.items():
        for filetype, filetyp_files in input_files.items():
            if get_alias(filetype) != which:
                continue
            # if a dir is given as a filepath, use all h5 files in that dir instead
            expanded_files = []
            for path in filetyp_files:
                if os.path.isdir(path):
                    expanded_files.extend(_get_h5_files(path))
                else:
                    expanded_files.append(path)
            files[input_name] = tuple(expanded_files)
            # store number of files for this output
            n_files.append(len(expanded_files))

    if n_files and n_files.count(n_files[0]) != len(n_files):
        raise ValueError(
            "Input with different number of {} in toml list".format(which))

    return files



from __future__ import division
import tensorflow as tf
import numpy as np


def conv4d(
        input_tensor,
        filters,
        kernel_size=(3, 3, 3, 3),
        strides=(1, 1, 1, 1),
        padding='valid',
        data_format='channels_first',
        dilation_rate=(1, 1, 1, 1),
        activation=None,
        use_bias=True,
        kernel_initializer=tf.ones_initializer(),
        bias_initializer=tf.zeros_initializer(),
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        trainable=True,
        name=None,
        reuse=None):
    """
    https://github.com/funkey/conv4d/blob/master/conv4d.py
    """

    # check arguments
    assert len(input_tensor.get_shape().as_list()) == 6, (
        "Tensor of shape (b, c, l, d, h, w) expected")
    assert len(kernel_size) == 4, "4D kernel size expected"
    assert strides == (1, 1, 1, 1), (
        "Strides other than 1 not yet implemented")
    assert data_format == 'channels_first', (
        "Data format other than 'channels_first' not yet implemented")
    assert dilation_rate == (1, 1, 1, 1), (
        "Dilation rate other than 1 not yet implemented")
    if not name:
        name = 'conv4d'

    # input, kernel, and output sizes
    (b, c_i, l_i, d_i, h_i, w_i) = tuple(input_tensor.get_shape().as_list())
    (l_k, d_k, h_k, w_k) = kernel_size

    # output size for 'valid' convolution
    if padding == 'valid':
        (l_o, d_o, h_o, w_o) = (
            l_i - l_k + 1,
            d_i - d_k + 1,
            h_i - h_k + 1,
            w_i - w_k + 1
        )
    else:
        (l_o, d_o, h_o, w_o) = (l_i, d_i, h_i, w_i)

    # output tensors for each 3D frame
    frame_results = [None]*l_o

    # convolve each kernel frame i with each input frame j
    for i in range(l_k):

        # reuse variables of previous 3D convolutions for the same kernel
        # frame (or if the user indicated to have all variables reused)
        reuse_kernel = reuse

        for j in range(l_i):

            # add results to this output frame
            out_frame = j - (i - l_k//2) - (l_i - l_o)//2
            if out_frame < 0 or out_frame >= l_o:
                continue

            # convolve input frame j with kernel frame i
            frame_conv3d = tf.layers.conv3d(
                tf.reshape(input_tensor[:,:,j,:], (b, c_i, d_i, h_i, w_i)),
                filters,
                kernel_size=(d_k, h_k, w_k),
                padding=padding,
                data_format='channels_first',
                activation=None,
                use_bias=use_bias,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                kernel_regularizer=kernel_regularizer,
                bias_regularizer=bias_regularizer,
                activity_regularizer=activity_regularizer,
                trainable=trainable,
                name=name + '_3dchan%d'%i,
                reuse=reuse_kernel)

            # subsequent frame convolutions should use the same kernel
            reuse_kernel = True

            if frame_results[out_frame] is None:
                frame_results[out_frame] = frame_conv3d
            else:
                frame_results[out_frame] += frame_conv3d

    output = tf.stack(frame_results, axis=2)

    if activation:
        output = activation(output)

    return output


class Conv4DBlock(tf.keras.Model):
    def __init__(self, **kwargs):
        super().__init__(name="")
        self.kwargs = kwargs

    def call(self, input_tensor):
        return conv4d(input_tensor=input_tensor, **self.kwargs)


def test_conv4d():
    i = np.round(np.random.random((1, 1, 10, 11, 12, 13))*100)
    input = tf.constant(i, dtype=tf.float32)
    bias_init = tf.constant_initializer(0)

    output = conv4d(
        input,
        1,
        (3, 3, 3, 3),
        data_format='channels_first',
        bias_initializer=bias_init,
        name='conv4d_valid')

    with tf.Session() as s:

        s.run(tf.global_variables_initializer())
        o = s.run(output)

        k0 = tf.get_default_graph().get_tensor_by_name(
            'conv4d_valid_3dchan0/kernel:0').eval().flatten()
        k1 = tf.get_default_graph().get_tensor_by_name(
            'conv4d_valid_3dchan1/kernel:0').eval().flatten()
        k2 = tf.get_default_graph().get_tensor_by_name(
            'conv4d_valid_3dchan2/kernel:0').eval().flatten()

        print("conv4d at (0, 0, 0, 0): %s"%o[0,0,0,0,0,0])
        i0 = i[0,0,0,0:3,0:3,0:3].flatten()
        i1 = i[0,0,1,0:3,0:3,0:3].flatten()
        i2 = i[0,0,2,0:3,0:3,0:3].flatten()

        compare = (i0*k0 + i1*k1 + i2*k2).sum()
        print("manually computed value at (0, 0, 0, 0): %s"%compare)

        print("conv4d at (4, 4, 4, 4): %s"%o[0,0,4,4,4,4])
        i0 = i[0,0,4,4:7,4:7,4:7].flatten()
        i1 = i[0,0,5,4:7,4:7,4:7].flatten()
        i2 = i[0,0,6,4:7,4:7,4:7].flatten()

        compare = (i0*k0 + i1*k1 + i2*k2).sum()
        print("manually computed value at (4, 4, 4, 4): %s"%compare)

    output = conv4d(
        input,
        1,
        (3, 3, 3, 3),
        data_format='channels_first',
        padding='same',
        kernel_initializer=tf.constant_initializer(1),
        bias_initializer=bias_init,
        name='conv4d_same')

    with tf.Session() as s:

        s.run(tf.global_variables_initializer())
        o = s.run(output)

        print("conv4d at (0, 0, 0, 0): %s"%o[0,0,0,0,0,0])
        i0 = i[0,0,0:2,0:2,0:2,0:2]
        print("manually computed value at (0, 0, 0, 0): %s"%i0.sum())

        print("conv4d at (5, 5, 5, 5): %s"%o[0,0,5,5,5,5])
        i5 = i[0,0,4:7,4:7,4:7,4:7]
        print("manually computed value at (5, 5, 5, 5): %s"%i5.sum())

        print("conv4d at (9, 10, 11, 12): %s"%o[0,0,9,10,11,12])
        i9 = i[0,0,8:,9:,10:,11:]
        print("manually computed value at (9, 10, 11, 12): %s"%i9.sum())


if __name__ == "__main__":
    test_conv4d()
